from __future__ import annotations

from collections import UserList
from dataclasses import dataclass, field
from dataclasses_json import dataclass_json, config
from typing import Optional, Union, List

import wxcadm
import wxcadm.location
import wxcadm.person
from wxcadm import log
from .common import *
from .models import OutboundProxy


class CallRouting:
    def __init__(self, org: wxcadm.Org):
        self.org = org
        """ The Org to which the Call Routing is associated """
        pass

    @property
    def trunks(self):
        """ The :py:class:`Trunks` instance for this Org """
        return Trunks(self.org)

    @property
    def route_groups(self):
        """ The :py:class:`RouteGroups` instance for this Org """
        return RouteGroups(self.org)

    @property
    def route_lists(self):
        """ The :py:class:`RouteLists` instance for this Org """
        return RouteLists(self.org)

    @property
    def dial_plans(self):
        """ The :py:class:`DialPlans` instance for this Org """
        return DialPlans(self.org)

    def test(self, originator: wxcadm.person.Person | Trunk,
             destination: str,
             orig_number: Optional[str] = None) -> dict:
        """ Test the Call Routing for a call.

        This method accepts the `originator` as either a :py:class:`Person` instance or a :py:class:`Trunk` instance,
        depending on whether the call is being placed by a user or arriving at Webex Calling over a Trunk.

        Args:
            originator (Person, Trunk): The Person or Trunk instance to test origination from
            destination (str): The destination number
            orig_number (str): For Trunk-originated calls, the originator number

        Returns:
            dict: Since the responses are dynamic, the full response dict from the Webex API is returned.

        """
        # First, figure out if the originator is a Person or a Trunk
        if isinstance(originator, wxcadm.person.Person):
            originator = originator.id
            orig_type = 'USER'
        elif isinstance(originator, Trunk):
            originator = originator.id
            orig_type = 'TRUNK'
        else:
            log.warning("CallRouting.test() called without a Person or Trunk instance")
            raise ValueError('originator argument must be a Person or Trunk instance')

        payload = {'originatorId': originator,
                   'originatorType': orig_type,
                   'destination': destination}
        if orig_type == 'TRUNK' and orig_number is not None:
            payload['originatorNumber'] = orig_number

        response = webex_api_call('post', '/v1/telephony/config/actions/testCallRouting/invoke',
                                  params={'orgId': self.org.id}, payload=payload)
        return response


class Trunks(UserList):
    """ Trunks is a class that behaves as an array. Each item in the array is a :py:class:`Trunk` instance."""
    def __init__(self, org: wxcadm.Org):
        log.info('Initializing Trunks instance')
        super().__init__()
        self.org = org
        self.data = []
        items = webex_api_call('get', '/v1/telephony/config/premisePstn/trunks')
        log.debug(f'Trunks from Webex: {items}')
        for item in items['trunks']:
            item['org'] = self.org
            this_trunk = Trunk.from_dict(item)
            self.data.append(this_trunk)

    def get(self, name: Optional[str] = None, id: Optional[str] = None) -> Optional[Trunk]:
        """ Get a Trunk by name or ID

        Args:
            name: The name of the Trunk
            id: The Trunk ID

        Returns:

        """
        if name is None and id is None:
            raise ValueError("Must provide name or id")
        for trunk in self.data:
            if trunk.id == id:
                return trunk
            if trunk.name == name:
                return trunk
        return None

    def add_trunk(self, name: str,
                  location: wxcadm.Location | str,
                  password: str,
                  dual_identity_support: bool,
                  type: str,
                  device_type: Optional[str] = 'Cisco Unified Border Element',
                  address: Optional[str] = None,
                  domain: Optional[str] = None,
                  port: Optional[int] = None,
                  max_concurrent_calls: Optional[int] = None):
        """ Create a Trunk for the organization.

        A Trunk is a connection between Webex Calling and the premises, which terminates on the premises with a local
        gateway or other supported device. The trunk can be assigned to a Route Group - a group of trunks that allow
        Webex Calling to distribute calls over multiple trunks or to provide redundancy.

        Args:
            name (str): A unique name for the Trunk
            location (Location): The Location instance to add the Trunk to. Also supports a Location ID as a string.
            password (str): The password to use with the Trunk
            dual_identity_support (bool): The Dual Identity Support setting impacts the handling of the From header and
                P-Asserted-Identity (PAI) header when sending an initial SIP INVITE to the trunk for an outbound call.
                When True, the From and PAI headers are treated independently and may differ. When False, the PAI header
                is set to the same value as the From header. Please refer to the documentation for more details.
            type (str): Either 'REGISTERING' or 'CERTIFICATE_BASED'
            device_type (str, optional): The type of device. Defaults to 'Cisco Unified Border Element'
            address (str, optional): FQDN or SRV address. Required to create a static certificate-based trunk.
            domain (str, optional): Domain name. Required to create a static certificate based trunk.
            port (int, optional): FQDN port. Required to create a static certificate-based trunk.
            max_concurrent_calls (int, optional): Max Concurrent call. Required to create a static certificate based
                trunk.

        Returns:
            bool: True on success, False otherwise

        """
        log.info(f'Adding Trunk: {name}')
        if isinstance(location, wxcadm.location.Location):
            location_id = location.id
        elif isinstance(location, str):
            location_id = location
        else:
            log.warning(f'Cannot determine Location ID with given location: {location}')
            return False

        payload = {'name': name,
                   'locationId': location_id,
                   'password': password,
                   'dualIdentitySupportEnabled': dual_identity_support,
                   'trunkType': type}
        if type.upper() == 'CERTIFICATE_BASED':
            payload['deviceType'] = device_type
            payload['address'] = address
            payload['domain'] = domain
            payload['port'] = port
            payload['maxConcurrentCalls'] = max_concurrent_calls

        success = webex_api_call('post', '/v1/telephony/config/premisePstn/trunks',
                                 params={'orgId': self.org.id},
                                 payload=payload)
        if success:
            return True
        else:
            return False


@dataclass_json
@dataclass
class Trunk:
    org: wxcadm.Org = field(repr=False)
    """ The Org to which the Trunk belongs """
    id: str
    """ The unique identifier of the Trunk """
    name: str
    """ The text name of the Trunk """
    location: dict | wxcadm.Location
    """ The Location instance associated with the Trunk """
    in_use: bool = field(metadata=config(field_name="inUse"))
    """ Whether the Trunk is in use """
    trunk_type: str = field(metadata=config(field_name="trunkType"))
    """ The type of Trunk, either 'REGISTERING' or 'CERTIFICATE_BASED' """
    dedicated_instance_only: bool = field(metadata=config(field_name="isRestrictedToDedicatedInstance"))
    """ Whether the Trunk can only be used by Webex Calling Dedicated Instance"""

    # Fields from a detail GET. They will be populated with a __getattr__ later
    otg_dtg: str = field(init=False, repr=False)
    """ The OTG/DTG value for the Trunk"""
    lineport: str = field(init=False, repr=False)
    """ The Line/Port identifier for the Trunk """
    used_by_locations: list[dict] = field(init=False, repr=False)
    """ List of Locations using the Trunk for PSTN routing """
    pilot_user_id: str = field(init=False, repr=False)
    """ The Pilot User ID for the Trunk """
    outbound_proxy: Optional[OutboundProxy] = field(init=False, repr=False)
    """ The :class:`~.models.OutboundProxy` for the Trunk """
    sip_auth_user: str = field(init=False, repr=False)
    """ The SIP Auth User ID for the Trunk """
    status: str = field(init=False, repr=False)
    """ The status of the Trunk """
    response_status: list = field(init=False, repr=False)
    """ List of status messages for the Trunk """
    dual_identity_support: bool = field(init=False, repr=False)
    """ Whether the Trunk supports Dual Identity """
    device_type: str = field(init=False, repr=False)
    """ The type of the device """
    max_calls: int = field(init=False, repr=False)
    """ The Max Concurrent Calls """

    def __post_init__(self):
        log.debug(f'Finding Location instance: {self.location}')
        # Since we only have a dict of the trunk location, go get the actual Location instance
        my_location = self.org.locations.get(id=self.location['id'])
        if my_location is not None:
            self.location = my_location

    def __getattr__(self, item):
        # The following is a crazy fix for a PyCharm debugger bug. It serves no purpose other than to stop extra API
        # calls when developing in PyCharm. See the following bug:
        # https://youtrack.jetbrains.com/issue/PY-48306
        if item == 'shape':
            return None
        log.debug(f"Collecting details for Trunk: {self.id}")
        response = webex_api_call('get', f"v1/telephony/config/premisePstn/trunks/{self.id}",
                                  params={'orgId': self.org.id})
        self.otg_dtg = response.get('otgDtgId', '')
        self.lineport = response.get('linePort', '')
        self.used_by_locations = response.get('locationsUsingTrunk', [])
        self.pilot_user_id = response.get('pilotUserId', '')
        proxy_dict = response.get('outboundProxy', None)
        if proxy_dict is not None:
            proxy = OutboundProxy.from_dict(proxy_dict)
            self.outbound_proxy = proxy
        else:
            self.outbound_proxy = None
        self.sip_auth_user = response.get('sipAuthenticationUserName', '')
        self.status = response.get('status', '')
        self.response_status = response.get('responseStatus', [])
        self.dual_identity_support = response.get('dualIdentitySupportEnabled', False)
        self.device_type = response.get('deviceType', '')
        self.max_calls = response.get('maxConcurrentCalls', 0)
        return self.__getattribute__(item)

    def set_dual_identity_support(self, enabled: bool):
        """ Set the Dual Identity Support flag for the Trunk

        The Dual Identity Support setting impacts the handling of the From header and P-Asserted-Identity (PAI) header
        when sending an initial SIP INVITE to the trunk for an outbound call. When enabled, the From and PAI headers
        are treated independently and may differ. When disabled, the PAI header is set to the same value as the From
        header. Please refer to the documentation for more details.

        Args:
            enabled (bool): Whether the Dual Identity Support flag is enabled or disabled

        Returns:
            bool: True on success, False otherwise

        Raises:
            wxcadm.ApiError: Raised when the command is rejected by the Webex API

        """
        log.debug(f'Setting Dual Identity Support on trunk: {self.id} to {enabled}')
        payload = {
            "dualIdentitySupportEnabled": enabled,
        }
        webex_api_call("put", f"v1/telephony/config/premisePstn/trunks/{self.id}",
                       payload=payload, params={'orgId': self.org.id})
        self.dual_identity_support = enabled
        return True

    def set_max_calls(self, calls: int):
        """ Sets the maximum number of Concurrent Calls for the Trunk

        .. note::
            Setting the maximum number of Concurrent Calls for the Trunk is only supported for certificate-based trunks.

        Args:
            calls (int): The maximum number of Concurrent Calls for the Trunk

        Returns:
            bool: True on success, False otherwise

        Raises:
            ValueError: Raised when trying to set the max calls value on a registration-based Trunk
            wxcadm.ApiError: Raised when the command is rejected by the Webex API

        """
        log.debug(f'Setting Max Concurrent Calls on trunk: {self.id} to {calls}')
        if self.trunkType == "REGISTERING":
            log.warning("Cannot set max_calls on REGISTERING Trunk")
            return ValueError("Cannot set max_calls on REGISTERING Trunk")
        payload = {
            "maxConcurrentCalls": calls
        }
        webex_api_call("put", f"v1/telephony/config/premisePstn/trunks/{self.id}",
                       payload=payload, params={'orgId': self.org.id})
        self.max_calls = calls
        return True


class RouteGroups(UserList):
    """ RouteGroups is a class that behaves as an array. Each item in the array is a :py:class:`RouteGroup` instance."""
    def __init__(self, org: wxcadm.Org):
        log.info('Initializing RouteGroups instance')
        super().__init__()
        self.org = org
        self.data = []
        items = webex_api_call('get', '/v1/telephony/config/premisePstn/routeGroups', params={'orgId': self.org.id})
        log.debug(f'Route Groups from Webex: {items}')
        for item in items['routeGroups']:
            this_rg = RouteGroup(self.org, **item)
            self.data.append(this_rg)

    def get(self, id: Optional[str] = None, name: Optional[str] = None) -> Optional[RouteGroup]:
        """ Return a RouteGroup instance with the given ID or name.

        Args:
            id (str, optional): The RouteGroup ID
            name (str, optional): The RouteGroup name

        Returns:
            RouteGroup: The matching RouteGroup instance. None is returned if no match is found.

        """
        for rg in self.data:
            if rg.id == id or rg.name == name:
                return rg
        return None


@dataclass
class RouteGroup:
    org: wxcadm.Org = field(repr=False)
    """ The Org to which the RouteGroup belongs """
    id: str
    """ The unique identifier for the RouteGroup """
    name: str
    """ The name of the RouteGroup """
    inUse: bool
    """ Whether or not the RouteGroup is being used by any Location, Route List or Dial Plan"""

    @property
    def trunks(self) -> List[dict]:
        """ A dict of the Trunks within the Route Group along with their Priorities """
        response = webex_api_call('get', f"v1/telephony/config/premisePstn/routeGroups/{self.id}",
                                  params={'orgId': self.org.org_id})
        return response['localGateways']

    def add_trunk(self, trunk: Trunk, priority: Union[str, int]):
        """ Add a new Trunk to the Route Group with the given priority

        The priority value will be a fixed number value (int) for the given priority. For a Route Group with an unknown
        number of existing Trunks, the keyword ``'next'`` can be used to assign the priority as the next numerical
        value, or the keyword ``'with_last'`` will assign the priority to the same value as other trunks with the last
        priority value.

        Args:
            trunk (Trunk): The :class:`Trunk` to add
            priority (str, int): Either a numeric priority or the word ``'next'`` or ``'with_last'``

        Returns:

        """
        trunks = self.trunks
        if priority == 'next' or priority =='with_last':
            highest_val = self.__get_last_priority(trunks)
            if priority == 'next':
                priority = int(highest_val) + 1
            else:
                priority = int(highest_val)

        trunks.append(
            {
                'id': trunk.id,
                'name': trunk.name,
                'locationId': trunk.location.id,
                'priority': priority
            }
        )
        payload = {
            'name': self.name,
            'localGateways': trunks
        }

        webex_api_call('put', f"v1/telephony/config/premisePstn/routeGroups/{self.id}",
                       params={'orgId': self.org.org_id}, payload=payload)
        return True

    def __get_last_priority(self, trunks: Optional[dict] = None):
        if trunks is None:
            trunks = self.trunks
        highest = 1
        for trunk in trunks:
            if trunk['priority'] > highest:
                highest = trunk['priority']
        return highest


class RouteLists(UserList):
    """ RouteLists is a class that behaves as an array. Each item in the array is a :py:class:`RouteList` instance."""
    # TODO - Create Route List, Delete Route List
    def __init__(self, org: wxcadm.Org):
        log.info('Initializing RouteLists instance')
        super().__init__()
        self.org = org
        self.data = []
        items = webex_api_call('get', '/v1/telephony/config/premisePstn/routeLists')
        log.debug(f'Route Lists from Webex: {items}')
        for item in items['routeLists']:
            this_rl = RouteList(self.org, **item)
            self.data.append(this_rl)

    def get(self, name: Optional[str] = None, id: Optional[str] = None) -> Optional[RouteLists]:
        """ Get a RouteList by name or ID

        Args:
            name: The name of the RouteList
            id: The RouteList ID

        Returns:
            RouteList

        """
        if name is None and id is None:
            raise ValueError("Must provide name or id")
        for routelist in self.data:
            if routelist.id == id:
                return routelist
            if routelist.name == name:
                return routelist
        return None


@dataclass
class RouteList:
    # TODO - Delete Route List
    org: wxcadm.Org = field(repr=False)
    """ The Org to which the RouteList belongs """
    id: str = field(repr=False)
    """ The unique identifier for the RouteList """
    name: str
    """ The name of the RouteList """
    locationId: str
    locationName: str
    routeGroupId: str
    routeGroupName: str

    def __post_init__(self):
        # This cleans up the Location and Route Group references so that they get the wxcadm instances for each
        self.route_group = self.org.call_routing.route_groups.get_route_group(id=self.routeGroupId)
        del self.routeGroupId, self.routeGroupName
        self.location = self.org.locations.get(id=self.locationId)
        del self.locationId, self.locationName

    @property
    def numbers(self):
        """ The numbers assigned to this RouteList """
        response = webex_api_call('get', f'/v1/telephony/config/premisePstn/routeLists/{self.id}/numbers')
        return response.json()


class DialPlans(UserList):
    """ DialPlans is a class that behaves as an array. Each item in the array is a :py:class:`DialPlan` instance."""
    def __init__(self, org: wxcadm.Org):
        log.info('Initializing DialPlans instance')
        super().__init__()
        self.org = org
        self.data = []
        items = webex_api_call('get', '/v1/telephony/config/premisePstn/dialPlans')
        log.debug(f'Dial Plans from Webex: {items}')
        for item in items['dialPlans']:
            this_dp = DialPlan(self.org, **item)
            self.data.append(this_dp)


@dataclass
class DialPlan:
    org: wxcadm.Org = field(repr=False)
    id: str = field(repr=False)
    name: str
    routeId: str
    routeName: str
    routeType: str

    _patterns: Optional[list] = field(init=False, default=None)

    @property
    def patterns(self):
        """ The Dial Patters within the DialPlan"""
        response = webex_api_call('get', f'/v1/telephony/config/premisePstn/dialPlans/{self.id}/dialPatterns')
        self._patterns = response['dialPatterns']
        return self._patterns

    def add_pattern(self, pattern: str) -> bool:
        """ Add a new dial pattern to the DialPlan

        Args:
            pattern (str): The pattern to add. This should be a valid Webex Dial Plan pattern.

        Returns:
            bool: True on success, False otherwise

        """
        payload = {'dialPatterns': [{'dialPattern': pattern, 'action': 'ADD'}]}
        success = webex_api_call('put', f'/v1/telephony/config/premisePstn/dialPlans/{self.id}/dialPatterns',
                                 payload=payload)
        if success:
            return True
        else:
            return False

    def delete_pattern(self, pattern: str) -> bool:
        """ Delete a dial pattern from the DialPlan

        Args:
            pattern (str): The pattern to delete. The pattern should already exist in the DialPlan

        Returns:
            bool: True on success, False otherwise

        """
        payload = {'dialPatterns': [{'dialPattern': pattern, 'action': 'DELETE'}]}
        success = webex_api_call('put', f'/v1/telephony/config/premisePstn/dialPlans/{self.id}/dialPatterns',
                                 payload=payload)
        if success:
            return True
        else:
            return False


class TranslationPattern:
    def __init__(self,
                 parent: Union[wxcadm.Org, wxcadm.Location],
                 config: Optional[dict] = None):
        log.info("Creating TranslationsPattern instance")
        self.parent = parent
        self.id: Optional[str] = None
        self.name: Optional[str] = None
        self.match_pattern: Optional[str] = None
        self.replacement_pattern: Optional[str] = None
        self.location: Optional[wxcadm.Location] = None
        self.__process_config(config)

    def __process_config(self, config: dict):
        log.debug(f"Processing config: {config}")
        self.id = config.get('id')
        self.name = config.get('name', '')
        self.match_pattern = config.get('matchingPattern')
        self.replacement_pattern = config.get('replacementPattern')
        self.level = config.get('level', None)
        if config.get('location', None) is not None:
            self.location = location_finder(config['location']['id'], self.parent)

    def update(self,
               name: Optional[str] = None,
               match_pattern: Optional[str] = None,
               replacement_pattern: Optional[str] = None):
        """ Update a Translation Pattern

        Args:
            name (str, optional): The new name of the Translation Pattern. Defaults to no change.
            match_pattern (str, optional): The new match pattern. Defaults to no change.
            replacement_pattern (str, optional): The new replacement pattern. Defaults to no change.

        Returns:
            bool: True on success

        """
        log.info(f"Updating TranslationPattern {self.id}")
        payload = {'name': name if name is not None else self.name,
                   'matchingPattern': match_pattern if match_pattern is not None else self.match_pattern,
                   'replacementPattern': replacement_pattern if replacement_pattern is not None \
                       else self.replacement_pattern}
        url = f'v1/telephony/config/callRouting/translationPatterns/{self.id}' if self.location is None \
            else f'v1/telephony/config/locations/{self.location.id}/callRouting/translationPatterns/{self.id}'
        response = webex_api_call('put', url, params={'orgId': self.parent.org_id}, payload=payload)
        log.debug(f"Response: {response}")
        return True

    def delete(self):
        """ Delete the Translation Pattern

        Returns:
            bool: True on success

        """
        log.info(f"Deleting Translation Pattern {self.id}")
        url = f'v1/telephony/config/callRouting/translationPatterns/{self.id}' if self.location is None \
            else f'v1/telephony/config/locations/{self.location.id}/callRouting/translationPatterns/{self.id}'
        webex_api_call('delete', url, params={'orgId': self.parent.org_id})
        return True


class TranslationPatternList(UserList):
    def __init__(self, parent: Union[wxcadm.Org, wxcadm.Location]):
        super().__init__()
        self.parent: Union[wxcadm.Org, wxcadm.Location] = parent
        self.data: list[wxcadm.TranslationPattern] = self._get_data()

    def _get_data(self):
        data = []
        params = {
            'orgId': self.parent.org_id
        }
        if isinstance(self.parent, wxcadm.Location):
            params['limitToLocationId'] = self.parent.id
        response = webex_api_call('get', "v1/telephony/config/callRouting/translationPatterns", params=params)
        for entry in response['translationPatterns']:
            data.append(TranslationPattern(self.parent, entry))
        return data

    def refresh(self):
        """ Refresh the list of Translations Patterns """
        self.data = self._get_data()
        return self.data

    def get(self,
            id: Optional[str] = None,
            name: Optional[str] = None,
            match_pattern: Optional[str] = None,
            replacement_pattern: Optional[str] = None,
            ):

        for entry in self.data:
            if entry.id == id or id is None:
                if entry.name == name or name is None:
                    if entry.match_pattern == match_pattern or match_pattern is None:
                        if entry.replacement_pattern == replacement_pattern or replacement_pattern is None:
                            return entry

        return None

    def create(self,
               name: str,
               match_pattern: str,
               replacement_pattern: str,
               location: Optional[wxcadm.Location] = None):
        """ Create a new Translation Pattern

        If this method is called for a :class:`TranslationPatternList` at the Org level, the ``location`` argument is
        optional, and, if omitted, the Translation Pattern will be created at the Org level. If the
        :class:`TranslationPatternList` is at the Location level, the ``location`` argument is still optional but will
        default to the Location of the list.

        Args:
            name (str): The name of the Translation Pattern
            match_pattern (str): The pattern to match
            replacement_pattern (str): The replacement pattern to apply to a match
            location (Location, optional): The :class:`Location` to create the Translation Pattern for

        Returns:
            TranslationPattern: The newly-created Translation Pattern.

        """
        log.info(f"Creating Translation Pattern. Name: {name}, Match: {match_pattern}, Replace: {replacement_pattern}")
        payload = {
            'name': name,
            'matchingPattern': match_pattern,
            'replacementPattern': replacement_pattern
        }
        scope = 'org'
        if location is not None:
            log.info(f"Location received in arg: {location.name}")
            url = f"v1/telephony/config/locations/{location.id}/callRouting/translationPatterns"
            scope = 'location'
            location_id = location.id
        else:
            if isinstance(self.parent, wxcadm.Location):
                log.info(f"No Location received. Using Location scope from parent: {self.parent.name}")
                url = f"v1/telephony/config/locations/{self.parent.id}/callRouting/translationPatterns"
                scope = 'location'
                location_id = self.parent.id
            else:
                log.info("No Location defined. Using Org scope.")
                url = f"v1/telephony/config/callRouting/translationPatterns"
                scope = 'org'

        response = webex_api_call('post', url, params={'orgId': self.parent.org_id}, payload=payload)
        pattern_id = response['id']
        log.debug(f"New Pattern ID: {pattern_id}")
        log.debug("Getting new pattern details")
        if scope == 'org':
            config = webex_api_call('get', f"v1/telephony/config/callRouting/translationPatterns/{pattern_id}",
                                    params={'orgId': self.parent.org_id})
            new_pattern = TranslationPattern(self.parent, config)
        else:
            config = webex_api_call(
                'get',
                f"v1/telephony/config/locations/{location_id}/callRouting/translationPatterns/{pattern_id}",
                params={'orgId': self.parent.org_id}
            )
            new_pattern = TranslationPattern(self.parent, config)
        return new_pattern




import unittest
import os
from dotenv import load_dotenv
import wxcadm
from random import choice
import queue
import time


class TestXSI(unittest.TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        load_dotenv()
        cls.access_token = os.getenv("WEBEX_ACCESS_TOKEN")
        if not cls.access_token:
            print("No WEBEX_ACCESS_TOKEN found. Cannot continue.")
            exit(1)

    def setUp(self) -> None:
        self.webex = wxcadm.Webex(self.access_token)
        endpoints = self.webex.org.get_xsi_endpoints()
        self.assertIsInstance(endpoints, dict)
        if endpoints is not None:
            self.xsi_capable: bool = True
        else:
            self.xsi_capable = False

    def test_person_profile(self):
        if self.xsi_capable is False:
            self.skipTest("Not an XSI Token")
        with self.subTest("My Profile"):
            me = self.webex.me
            me.start_xsi()
            my_profile = me.xsi.profile
            self.assertIsInstance(my_profile, dict)
        with self.subTest('Random Person Profile'):
            person = choice(self.webex.org.people.webex_calling())
            person.start_xsi()
            profile = person.xsi.profile
            self.assertIsInstance(profile, dict)

    def test_enterprise_events(self):
        if self.xsi_capable is False:
            self.skipTest("Not an XSI Token")
        events = wxcadm.XSIEvents(self.webex.org)
        self.assertIsInstance(events, wxcadm.xsi.XSIEvents)
        events_queue = queue.Queue()
        channel = events.open_channel(events_queue)
        self.assertIsInstance(channel, wxcadm.xsi.XSIEventsChannelSet)
        time.sleep(5)
        subscription = channel.subscribe("Advanced Call")
        if subscription is not False:
            self.assertIsInstance(subscription, wxcadm.xsi.XSIEventsSubscription)
            self.assertEqual(subscription.active, True)
            if subscription.active is True:
                channel.unsubscribe("all")
                message = events_queue.get()
                self.assertIsInstance(message, dict)
        channel.close()
        self.assertEqual(len(channel.subscriptions), 0)
        self.assertEqual(len(channel.active_channels), 0)

    def test_person_events(self):
        if self.xsi_capable is False:
            self.skipTest("Not an XSI Token")
        events = wxcadm.XSIEvents(self.webex.org)
        self.assertIsInstance(events, wxcadm.xsi.XSIEvents)
        events_queue = queue.Queue()
        channel = events.open_channel(events_queue)
        self.assertIsInstance(channel, wxcadm.xsi.XSIEventsChannelSet)
        time.sleep(5)
        subscription = channel.subscribe("Advanced Call", person=choice(self.webex.org.people.webex_calling()))
        self.assertIsInstance(subscription, wxcadm.xsi.XSIEventsSubscription)
        self.assertEqual(subscription.active, True)
        if subscription.active is True:
            channel.unsubscribe("all")
            message = events_queue.get()
            self.assertIsInstance(message, dict)
        channel.close()
        self.assertEqual(len(channel.subscriptions), 0)
        self.assertEqual(len(channel.active_channels), 0)

    def test_directory_all(self):
        if self.xsi_capable is False:
            self.skipTest("Not an XSI Token")
        me = self.webex.me
        self.assertIsInstance(me, wxcadm.person.Me)
        me.start_xsi()
        with self.subTest('Full Directory'):
            directory = me.xsi.directory()
            self.assertGreaterEqual(len(directory), 1)
        with self.subTest("Group Directory"):
            directory = me.xsi.directory('Group')
            self.assertGreaterEqual(len(directory), 1)
        with self.subTest("Name Search"):
            person = choice(self.webex.org.people.webex_calling())
            directory = me.xsi.directory(first_name=person.first_name)
            self.assertGreaterEqual(len(directory), 1)
        with self.subTest("Name Search - Case Insensitive"):
            person = choice(self.webex.org.people.webex_calling())
            search_string = f"{person.first_name.lower()}/i"
            directory = me.xsi.directory(first_name=search_string)
            self.assertGreaterEqual(len(directory), 1)
        with self.subTest("Any Match Search"):
            person1 = choice(self.webex.org.people.webex_calling())
            person2 = choice(self.webex.org.people.webex_calling())
            directory = me.xsi.directory(first_name=person1.first_name, last_name=person2.last_name, any_match=True)
            self.assertGreaterEqual(len(directory), 2)
        with self.subTest("Personal Directory"):
            directory = me.xsi.directory("personal")
            self.assertIsInstance(directory, list)


if __name__ == '__main__':
    unittest.main()



from __future__ import annotations

import requests
import wxcadm
from wxcadm import log
from .common import *
from .exceptions import *


class Terminus:
    """The Terminus class handles API calls using the Terminus API."""

    def __init__(self, org: wxcadm.Org, access_token: str):
        self._parent = org
        self._access_token = access_token
        self._headers = {"Authorization": f"Bearer {access_token}"}

        # Calculate the "customer" ID from the Org ID
        spark_id = decode_spark_id(org.id)
        self._customer = spark_id.split("/")[-1]

        self._url_base = f"https://terminus.huron-dev.com/api/v2/customers/{self._customer}/"
        self._server = "https://terminus.huron-dev.com"

    def get_locations(self):
        log.info("Getting locations from Terminus")
        r = requests.get(self._url_base + "locations", headers=self._headers)
        if r.ok:
            locations = r.json()
            for location in locations:
                webex_location = self._parent.locations.get(name=location['name'])
                webex_location.terminus_config = location
            return locations
        else:
            raise APIError("Your API Access Token does not have permission to get the locations from Terminus.")



import unittest
import os
from dotenv import load_dotenv
import wxcadm
from random import choice


class TestWebex(unittest.TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        load_dotenv()
        cls.access_token = os.getenv("WEBEX_ACCESS_TOKEN")
        if not cls.access_token:
            print("No WEBEX_ACCESS_TOKEN found. Cannot continue.")
            exit(1)

    def setUp(self) -> None:
        self.webex = wxcadm.Webex(self.access_token)
        self.random_location = choice(self.webex.org.locations.webex_calling())

    def test_bad_token(self):
        with self.assertRaises(wxcadm.TokenError):
            wxcadm.Webex("Bad Token Value")

    def test_org_count(self):
        org_count = len(self.webex.orgs)
        self.assertGreater(org_count, 0)

    def test_location_list(self):
        self.assertIsInstance(self.webex.org.locations, wxcadm.location.LocationList)
        self.assertIsInstance(self.random_location, wxcadm.location.Location)

    def test_location_hunt_groups(self):
        location = self.random_location
        self.assertIsInstance(location.hunt_groups, wxcadm.hunt_group.HuntGroupList)
        if location.hunt_groups:
            self.assertIsInstance(choice(location.hunt_groups), wxcadm.hunt_group.HuntGroup)

    # def test_location_create_update_delete(self):
    #     self.webex.org.locations: wxcadm.location.LocationList
    #     new_loc
    #     print(new_loc)
    #     self.assertIsInstance(new_loc, dict)

    def test_location_call_queues(self):
        location = self.random_location
        self.assertIsInstance(location.call_queues, wxcadm.call_queue.CallQueueList)
        if location.call_queues:
            self.assertIsInstance(choice(location.call_queues), wxcadm.call_queue.CallQueue)

if __name__ == '__main__':
    unittest.main()



from __future__ import annotations

import requests
from requests_toolbelt import MultipartEncoder
import base64
import os
from typing import Optional, Union
from dataclasses import dataclass, field
from collections import UserList

import wxcadm.exceptions
from .common import *
from .xsi import XSI
from .device import DeviceList
from .location import Location
from .monitoring import MonitoringList

from wxcadm import log


class PersonList(UserList):
    def __init__(self, parent: Union[wxcadm.Org, wxcadm.Location]):
        super().__init__()
        log.debug("Initializing PersonList")
        self.parent: Union[wxcadm.Org, wxcadm.Location] = parent
        self.data: list = self._get_people()

    def _get_people(self) -> list[Person]:
        log.debug("_get_people() started")
        params = {"callingData": "true"}

        if isinstance(self.parent, wxcadm.Org):
            log.debug(f"Using Org ID {self.parent.id} as Person filter")
            params['orgId'] = self.parent.id
        elif isinstance(self.parent, wxcadm.Location):
            log.debug(f"Using Location ID {self.parent.id} as Person filter")
            params['locationId'] = self.parent.id
            params['orgId'] = self.parent.org_id
        else:
            log.warn("Parent class is not Org or Location, so all People will be returned")
        response = webex_api_call("get", "v1/people", params=params)
        log.info(f"Found {len(response)} People")

        people = []

        for entry in response:
            people.append(Person(entry['id'], parent=self.parent, config=entry))
        return people

    def refresh(self):
        """ Refresh the list of :py:class:`Person` instances from Webex

        Returns:
            bool: True on success, False otherwise.

        """
        self.data = self._get_people()

    def get_by_id(self, id: str) -> Optional[Person]:
        """ Get the :py:class:`Person` with the given Person ID

        Args:
            id (str): The Person ID to find

        Returns:
            Person: The :py:class:`Person` instance. None is return if no match is found.

        """
        entry: Person
        for entry in self.data:
            if entry.id == id:
                return entry
        return None

    def get(self, id: Optional[str] = None, email: Optional[str] = None, name: Optional[str] = None,
            location: Optional[wxcadm.Location] = None):
        """ Get the :py:class:`Person` (or list) that matches the provided arguments

        This method was added after the :meth:`get_by_email()` and :meth:`get_by_id()` to match other List Classes.
        When the method is called with a ``name`` argument, a list *can* be returned if more than one Person matches
        the argument. Only the :attr:`Person.display_name` is checked for ``name`` matches. Name matches are
        case-insensitive and must match the entire Display Name.

        Args:
            id (str, optional): The Webex ID of the Person
            email (str, optional): The email address of the Person
            name (str, optional): The Display Name of the Person
            location (Location, optional): Return people at the specified Location

        Returns:
            Person: The :class:`Person` instance. None is returned if no match is found. If multiple matches are found
                for a ``name`` or ``location``, a list of :class:`Person` instances is returned.

        """
        if id is not None:
            return self.get_by_id(id)
        if email is not None:
            return self.get_by_email(email)
        if name is not None:
            entry: Person
            matches = []
            for entry in self.data:
                if entry.display_name.lower() == name.lower():
                    matches.append(entry)
            if len(matches) == 0:
                return None
            elif len(matches) == 1:
                return matches[0]
            else:
                return matches
        if location is not None:
            if not isinstance(location, wxcadm.Location):
                location = self.parent.locations.get(id=location)
                if location is None:
                    raise ValueError("Invalid Location ID")
            entry: Person
            matches = []
            for entry in self.data:
                if entry.location == location.id:
                    matches.append(entry)
            return matches
        raise KeyError("No valid search arguments provided")

    def get_by_email(self, email: str) -> Optional[Person]:
        """ Get the :py:class:`Person` with the given email address

        Args:
            email (str): The email address to find

        Returns:
            Person: The :py:class:`Person` instance. None is returned if no match is found.

        """
        entry: Person
        for entry in self.data:
            if entry.email.lower() == email.lower():
                return entry
        return None

    def webex_calling(self, enabled: bool = True) -> list[Person]:
        """ Return a list of :py:class:`Person` where Webex Calling is enabled/disabled

        Args:
            enabled (bool, optional): True (default) returns Webex Calling people. False returns people without
                Webex Calling

        Returns:
            list[:py:class:`Person`]: List of :py:class:`Person` instances. An empty list is returned if none match the
                given criteria

        """
        people = []
        entry: Person
        for entry in self.data:
            if entry.wxc is enabled:
                people.append(entry)
        return people

    def recorded(self, enabled: bool = True) -> list[Person]:
        """ Return a list of :py:class:`Person` where Call Recording is enabled/disabled

        Args:
            enabled (bool, optional): True (default) returns Recorded people. False returns people without Recording.

        Returns:
            list[:py:class:`Person`]: List of :py:class:`Person` instances. An empty list is returned if none match the
                given criteria.

        """
        people = []
        for entry in self.webex_calling(True):
            rec_config = entry.get_call_recording()
            if rec_config['enabled'] is enabled:
                people.append(entry)
        return people

    def create(self, email: str,
               location: Optional[Union[str, Location]] = None,
               licenses: list = None,
               calling: bool = True,
               phone_number: str = None,
               extension: str = None,
               first_name: str = None,
               last_name: str = None,
               display_name: str = None,
               ):
        """ Create a new user in Webex.

        Args:
            email (str): The email address of the user
            location (Location): The Location instance to assign the user to. Also accepts the Location ID as a string.
                This argument is not needed when the PeopleList parent is an Org.
            licenses (list, optional): List of license IDs to assign to the user. Use this when the license IDs
                are known. To have the license IDs determined dynamically, use the `calling`, `messaging` and
                meetings` parameters.
            calling (bool, optional): BETA - Whether to assign Calling licenses to the user. Defaults to True.
            phone_number (str, optional): The phone number to assign to the user.
            extension (str, optional): The extension to assign to the user
            first_name (str, optional): The user's first name. Defaults to empty string.
            last_name (str, optional): The users' last name. Defaults to empty string.
            display_name (str, optional): The full name of the user as displayed in Webex. If first name and last name
                are passed without display_name, the display name will be the concatenation of first and last name.

        Returns:
            Person: The Person instance of the newly-created user.

        """
        log.info(f"Creating new user: {email}")
        if (first_name or last_name) and display_name is None:
            log.debug("No display_name provided. Setting default.")
            display_name = f"{first_name} {last_name}"
        elif (first_name is None and last_name is None) and display_name is None:
            log.debug("No names provided. Using email as display_name")
            display_name = email

        # Find the license IDs for each requested service, unless licenses was passed
        if not licenses:
            log.debug("No licenses specified. Finding licenses.")
            licenses = []
            if calling:
                log.debug("Calling requested. Finding Calling licenses.")
                if isinstance(self.parent, wxcadm.Org):
                    calling_license = self.parent.get_wxc_person_license()
                elif isinstance(self.parent, wxcadm.Location):
                    calling_license = self.parent.parent.get_wxc_person_license()
                else:
                    raise wxcadm.exceptions.LicenseError("No Calling Licenses found")
                log.debug(f"Using Calling License: {calling_license}")
                licenses.append(calling_license)

        # Build the payload to send to the API
        log.debug("Building payload.")
        if isinstance(self.parent, wxcadm.Org):
            if isinstance(location, Location):
                location_id = location.id
            else:
                location_id = location
            org_id = self.parent.id
        elif isinstance(self.parent, wxcadm.Location):
            location_id = self.parent.id
            org_id = self.parent.parent.id
        else:
            raise ValueError("Unknown parent instance type")

        payload = {"emails": [email], "locationId": location_id, "orgId": org_id, "licenses": licenses}
        if phone_number is not None:
            payload["phoneNumbers"] = [{"type": "work", "value": phone_number}]
        if extension is not None:
            payload["extension"] = extension
        if first_name is not None:
            payload["firstName"] = first_name
        if last_name is not None:
            payload["lastName"] = last_name
        if display_name is not None:
            payload["displayName"] = display_name
        log.debug(f"Payload: {payload}")
        response = webex_api_call("post", "v1/people", params={'callingData': "true"}, payload=payload)
        if response:
            new_person = Person(response['id'], self.parent, response)
            return new_person
        else:
            raise wxcadm.exceptions.PutError("Something went wrong while creating the user")


class Person:
    def __init__(self, user_id, parent: Union[wxcadm.Org, wxcadm.Location] = None, config: dict = None):
        """ Initialize a new Person instance.

        If only the `user_id` is provided, the API calls will be made to get
        the config from Webex. To save on API calls, the config can be provided which will set the attributes
        without an API call.

        Args:
            user_id (str): The Webex ID of the person
            parent (object, optional): The parent object that created the Person instance. Used when the Person
                is created within the Org instance
            config (dict, optional): A dictionary of raw values from the `GET v1/people` items. Not normally used
                except for automated people population from the Org init.

        """
        self.id = user_id
        """The Webex ID of the Person"""
        self._parent = parent
        """The parent instance that created this Person"""
        # Attributes
        self.email: str = ""
        """The user's email address"""
        self.first_name: str = ""
        """The user's first name"""
        self.last_name: str = ""
        """The user's last name"""
        self.display_name: str = ""
        """The user's name as displayed in Webex"""
        self.wxc: bool = False
        '''True if this is a Webex Calling User'''
        self.licenses: list = []
        """List of licenses assigned to the person"""
        self.location: str = ""
        """The Webex ID of the user's assigned location"""
        self.roles: list = []
        """ The role IDs assigned to this Person in Webex"""
        self.vm_config: dict = {}
        '''Dictionary of the VM config as returned by Webex API with :meth:`get_vm_config()`'''
        self.call_recording: dict = {}
        """Dictionary of the Recording config as returned by Webex API with :meth:`get_call_recording()`"""
        self.barge_in: dict = {}
        """Dictionary of Barge-In config as returned by Webex API with :meth:`get_barge_in`"""
        self.call_forwarding: dict = {}
        """Dictionary of the Call Forwarding config as returned by Webex API 
        with :meth:`get_call_forwarding()`"""
        self.caller_id: dict = {}
        """Dictionary of Caller ID config as returned by Webex API with :meth:`get_caller_id()`"""
        self.intercept: dict = {}
        """Dictionary of Call Intercept config as returned by Webex API with :meth:`get_intercept()`"""
        self.dnd: dict = {}
        """Dictionary of DND settings as returned by Webex API with :meth:`get_dnd()`"""
        self.calling_behavior: dict = {}
        """Dictionary of Calling Behavior as returned by Webex API with :meth:`get_calling_behavior()`"""
        self.hoteling: dict = {}
        """Dictionary of Hoteling settings as returned by Webex API with :meth:`get_hoteling()`"""
        self.ptt: Optional[dict] = None
        """ Dictionary of Push-to-Talk settings as returned by Webex API with :meth:`get_ptt()` """
        self.xsi = None
        """Holds the XSI instance when created with the :meth:`start_xsi()` method."""
        self.numbers: list = []
        """The phone numbers for this person from Webex CI"""
        self.extension: Optional[str] = None
        """The extension for this person"""
        self._hunt_groups: list = []
        """A list of the Hunt Group instances that this user is an Agent for"""
        self._call_queues: list = []
        """A list of the Call Queue instances that this user is an Agent for"""
        self.outgoing_permission: dict = {}
        """Dictionary of Outgoing Permission config returned by Webex API
        with :meth:`get_outgoing_permission()`"""
        self.applications_settings = None
        """ The Application Services Settings for this Person"""
        self.executive_assistant = None
        self.avatar: Optional[str] = None
        """ The URL of the Person's avatar """
        self.department: Optional[str] = None
        """ The department the Person belongs to """
        self.manager: Optional[str] = None
        """ The Person's manager """
        self.login_enabled: Optional[bool] = None
        """ Whether the person is allowed to sign in """
        self.manager_id: Optional[str] = None
        """ The Person ID of the manager """
        self.title: Optional[str] = None
        """ The Person's title """
        self.addresses: Optional[list] = None
        """ A list of addresses for the Person """
        self.status: Optional[str] = None
        """ The current presence status of the Person """
        self._devices: Optional[DeviceList] = None
        self._monitoring: Optional[MonitoringList] = None

        # API-related attributes
        self._headers = parent._headers
        self._params = {"orgId": parent.id, "callingData": "true"}

        # If the config was passed, process it. If not, make the API call for the Person ID and then process
        if config:
            self.__process_api_data(config)
        else:
            response = self.__get_webex_data(f"v1/people/{self.id}")
            self.__process_api_data(response)

    def __process_api_data(self, data: dict):
        """Takes the API data passed as the `data` argument and parses it to the instance attributes.

        Args:
            data (dict): A dictionary of the raw data returned by the `v1/people` API call

        """
        self.email = data.get('emails', [''])[0]
        self.extension = data.get("extension", "")
        self.location = data.get("locationId", "")
        self.display_name = data.get("displayName", "")
        self.first_name = data.get("firstName", "")
        self.last_name = data.get("lastName", "")
        self.avatar = data.get('avatar', None)
        self.department = data.get('department', None)
        self.manager = data.get('manager', None)
        self.manager_id = data.get('managerId', None)
        self.title = data.get('title', None)
        self.addresses = data.get('addresses', None)
        self.status = data.get('status', None)
        self.login_enabled = data.get('loginEnabled', None)
        self.roles = data.get("roles", [])
        self.numbers = data.get("phoneNumbers", [])
        self.licenses = data.get("licenses", [])
        for license in self.licenses:
            if license in self._parent.wxc_licenses:
                self.wxc = True

    def __str__(self):
        return f"{self.email},{self.display_name}"

    def __repr__(self):
        return self.id

    # The following is to simplify the API call. Eventually I may open this as a public method to
    # allow arbitrary API calls
    def __get_webex_data(self, endpoint: str, params: dict = None):
        """ Issue a GET to the Webex API

        Args:
            endpoint (str): The endpoint of the call (i.e. "v1/people" or "/v1/people/{Person.id}")
            params (dict): Any additional params to be passed in the query (i.e. {"callingData":"true"}

        Returns:
            dict: The response from the Webex API

        """
        if params is None:
            params = {}
        log.debug(f"__get_webex_data started using {endpoint}")
        my_params = {**params, **self._params}
        r = requests.get(_url_base + endpoint, headers=self._headers, params=my_params)
        if r.status_code in [200]:
            response = r.json()
            return response
        else:
            return False

    def __put_webex_data(self, endpoint: str, payload: dict, params: Optional[dict] = None):
        """Issue a PUT to the Webex API

        Args:
            endpoint: The endpoint of the call (i.e. "v1/people" or "/v1/people/{Person.id}")
            payload: A dict to send as the JSON payload of the PUT
            params: Any additional params to be passed in the query (i.e. {"callingData":"true"}

        Returns:
            bool: True if successful, False if not

        """
        if params is None:
            params = {}
        log.debug(f"__put_webex_data started using {endpoint}")
        my_params = {**params, **self._params}
        log.debug(f"Params: {params}")
        log.debug(f"Payload: {payload}")
        r = requests.put(_url_base + endpoint, headers=self._headers, params=my_params, json=payload)
        if r.ok:
            log.info("Push successful")
            return True
        else:
            log.warning("__put_webex_data() failed")
            log.debug(f"[{r.status_code}] {r.text}")
            return False

    @property
    def org_id(self) -> Optional[str]:
        """ The Org ID for the Person """
        if isinstance(self._parent, wxcadm.Org):
            return self._parent.org_id
        elif isinstance(self._parent, wxcadm.Location):
            return self._parent.org_id
        else:
            return None

    @property
    def spark_id(self):
        """ The internal identifier used within Webex """
        user_id_bytes = base64.b64decode(self.id + "===")
        spark_id = user_id_bytes.decode("utf-8")
        return spark_id

    @property
    def name(self):
        """ The name of the Person. This is an alias for :attr:`display_name` and will return the same value """
        return self.display_name

    def role_names(self):
        """ Returns a list of the user's Roles, using the Role Name rather than the Role ID

        Returns:
            list: A list of Role names. None is returned if the Person has no administrative Roles

        """
        if len(self.roles) > 0:
            roles = []
            for role in self.roles:
                roles.append(self._parent.roles[role])
                return roles
            else:
                return None

    def assign_wxc(self,
                   location: wxcadm.Location,
                   phone_number: Optional[str] = None,
                   extension: Optional[str] = None,
                   unassign_ucm: Optional[bool] = False,
                   license_type: Optional[str] = 'Professional'):
        """ Assign Webex Calling to the user, along with a phone number and/or an extension.

        Args:
            location (Location): The Location instance to assign the Person to
            phone_number (str, optional): The phone number to assign to the Person. Defaults to None
            extension (str, optional): The extension to assign to the Person. Defaults to None
            unassign_ucm (bool, optional): True if you also want to remove the UCM license for the user. Default: False
            license_type (str, optional): 'Standard' or 'Professional'. Defaults to 'Professional'.

        Returns:
            bool: True on success, False if otherwise

        """
        if license_type.upper() == 'PROFESSIONAL':
            # To assign Webex Calling to a Person, we need to find the License ID for Webex Calling Professional
            license = self._parent.get_wxc_person_license()
            self.licenses.append(license)
        elif license_type.upper() == 'STANDARD':
            license = self._parent.get_wxc_standard_license()
            self.licenses.append((license))

        if unassign_ucm is True:
            # Figure out what the licenses are for UCM
            ucm_licenses = []
            for license in self._parent.licenses:
                if license['name'] == 'Unified Communication Manager (UCM)':
                    ucm_licenses.append(license['id'])
            # Then remove them from the user
            for license in self.licenses[:]:
                if license in ucm_licenses:
                    self.licenses.remove(license)

        # Call the update_person() method to update the new values.
        if phone_number is None:
            success = self.update_person(extension=extension, location=location.id)
        else:
            success = self.update_person(numbers=[{"type": "work", "value": phone_number}],
                                         extension=extension, location=location.id)
        if success:
            return True
        else:
            return False

    def unassign_wxc(self) -> bool:

        return True

    def start_xsi(self, get_profile: bool = False, cache: bool = False):
        """Starts an XSI session for the Person

        Args:
            get_profile (bool, optional): Whether to automatically get the XSI profile for the Person. Defaults to False
            cache (bool, optional): Whether to cache results so that the data doesn't need to be re-pulled from Webex.
                Defaults to False.

        Returns:
            XSI: The XSI instance for this Person
        """
        self.xsi = XSI(self, get_profile=get_profile, cache=cache)
        return self.xsi

    def reset_vm_pin(self, pin: str = None):
        """Resets the user's voicemail PIN.

        If no PIN is provided, the reset command is sent and the temporary PIN will be provided to the user via email.

        Args:
            pin (str, optional): The new temporary PIN to set for the Person

        Returns:
            bool: True on success, False otherwise.

        """
        log.info(f"Resetting VM PIN for {self.email}")
        if pin is not None:
            webex_api_call('put', f'v1/telephony/config/people/{self.id}/voicemail/passcode',
                           payload={'passcode': pin})
        else:
            webex_api_call("post", f"v1/people/{self.id}/features/voicemail/actions/resetPin/invoke",
                           params={"orgId": self._parent.id})
        return True

    def get_full_config(self):
        """
        Fetches all Webex Calling settings for the Person. Due to the number of API calls, this
        method is not performed automatically on Person init and should be called for each Person during
        any subsequent processing. If you are only interested in one of the features, calling that method
        directly can significantly improve the time to return data.

        """
        log.info(f"Getting the full config for {self.email}")
        if self.wxc:
            self.get_call_forwarding()
            self.get_vm_config()
            self.get_caller_id()
            self.get_call_recording()
            self.get_dnd()
            self.get_calling_behavior()
            self.get_caller_id()
            self.get_hoteling()
            self.get_barge_in()
            self.get_intercept()
            self.get_outgoing_permission()
            self.get_ptt()
            return self
        else:
            log.info(f"{self.email} is not a Webex Calling user.")

    @property
    def user_groups(self):
        """ List of the :py:class:`UserGroup` that the Person is assigned to

        Returns:
            list[UserGroup]: A list of :py:class:`UserGroup`s.

        """
        return self._parent.usergroups.find_person_assignments(self)

    @property
    def devices(self):
        """ :class:`DeviceList` of :py:class:`Device` instances for the Person

        Returns:
            DeviceList

        """
        if self._devices is None:
            self._devices = DeviceList(self)
        return self._devices

    @property
    def hunt_groups(self):
        """The Hunt Groups that this user is an Agent for.

        Returns:
            list[HuntGroup]: A list of the `HuntGroup` instances the user belongs to

        """
        log.info(f"Getting Hunt Groups for {self.email}")
        hunt_groups = []
        for hg in self._parent.hunt_groups:
            # Step through the agents for the Hunt Group to see if this person is there
            for agent in hg.agents:
                if agent['id'] == self.id:
                    hunt_groups.append(hg)
        self._hunt_groups = hunt_groups
        return self._hunt_groups

    @property
    def call_queues(self):
        """The Call Queues that this user is an Agent for.

        Returns:
            list[CallQueue]: A list of the `CallQueue` instances the user belongs to

        """
        log.info(f"Getting Hunt Groups for {self.email}")
        call_queues = []
        for cq in self._parent.call_queues:
            # Step through the agents for the Hunt Group to see if this person is there
            for agent in cq.config['agents']:
                if agent['id'] == self.id:
                    call_queues.append(cq)
        self._call_queues = call_queues
        return self._call_queues

    def get_call_forwarding(self):
        """Get the Call Forwarding config for the Person

        Returns:
            dict: The Call Forwarding config for the Person instance
        """
        log.info(f"Getting Call Forwarding config for {self.email}")
        self.call_forwarding = self.__get_webex_data(f"v1/people/{self.id}/features/callForwarding")
        return self.call_forwarding

    def get_barge_in(self):
        """Get the Barge-In config for the Person

        Returns:
            dict: The Barge-In config for the Person instance
        """
        log.info(f"Getting Barge-In config for {self.email}")
        self.barge_in = self.__get_webex_data(f"v1/people/{self.id}/features/bargeIn")
        return self.barge_in

    def push_barge_in(self, config: dict):
        """ Push the Barge-In config to Webex

        Returns:
            bool: True on success, False otherwise

        """
        log.info(f"Pushing Barge-In config for {self.email}")
        success = self.__put_webex_data(f"v1/people/{self.id}/features/bargeIn", payload=config)
        if success:
            return True
        else:
            log.warning(f"The Barge-In config push failed")
            return False

    def get_vm_config(self):
        """Fetch the Voicemail config for the Person from the Webex API"""
        log.info(f"Getting VM config for {self.email}")
        self.vm_config = self.__get_webex_data(f"v1/people/{self.id}/features/voicemail")
        return self.vm_config

    def push_vm_config(self, vm_config: dict = None):
        """ Push the Voicemail config back to Webex

        If the vm_config dict is provided, it will be sent as the payload to Webex. If it is omitted, the current
        Person.call_forwarding attributes will be sent.

        Args:
            vm_config (dict, optional: The vm_config dictionary to push to Webex

        Returns:
            dict: The new config that was sent back by Webex. False is returned if the API call fails.

        """
        log.info(f"Pushing VM Config for {self.email}")
        if vm_config is not None:
            payload = vm_config
        else:
            payload = self.vm_config
        success = self.__put_webex_data(f"v1/people/{self.id}/features/voicemail", payload)
        if success:
            self.get_vm_config()
            return self.vm_config
        else:
            return False

    def upload_busy_greeting(self, filename: str, activate: bool = True):
        """ Upload a WAV file to be used as the Person's Voicemail Busy Greeting

        This method does not check the format of the WAV file or provide any media conversion. The file should be in
        the correct format to be accepted by Webex. Files in the wrong format will fail and the method will return
        False.

        Args:
            filename (str): The filename (and path, if needed) of the WAV file to upload
            activate (bool, optional): Whether to activate this as the current Busy greeting. Defaults to True

        Returns:
            bool: True on success, False otherwise

        """
        log.info(f"Uploading VB Busy Greeting for {self.email}")
        upload_as = os.path.basename(filename)
        content = open(filename, "rb")
        encoder = MultipartEncoder(fields={"file": (upload_as, content, 'audio/wav')})
        log.debug(f"File Encoder: {encoder}")
        r = requests.post(_url_base + f"v1/people/{self.id}/features/voicemail/actions/uploadBusyGreeting/invoke",
                          headers={"Content-Type": encoder.content_type, **self._headers},
                          data=encoder)
        content.close()
        if r.ok:
            if activate is True:
                vm_config = self.get_vm_config()
                vm_config['sendBusyCalls']['greeting'] = "CUSTOM"
                self.push_vm_config(vm_config)
            return True
        else:
            log.warning(f"The Greeting upload failed: {r.text}")
            return False

    def upload_no_answer_greeting(self, filename: str, activate: bool = True):
        """ Upload a WAV file to be used as the Person's Voicemail No Answer Greeting

        This method does not check the format of the WAV file or provide any media conversion. The file should be in
        the correct format to be accepted by Webex. Files in the wrong format will fail and the method will return
        False.

        Args:
            filename (str): The filename (and path, if needed) of the WAV file to upload
            activate (bool, optional): Whether to activate this as the current No Answer greeting. Defaults to True

        Returns:
            bool: True on success, False otherwise

        """
        log.info(f"Uploading VB No Answer Greeting for {self.email}")
        upload_as = os.path.basename(filename)
        content = open(filename, "rb")
        encoder = MultipartEncoder(fields={"file": (upload_as, content, 'audio/wav')})
        log.debug(f"File Encoder: {encoder}")
        r = requests.post(_url_base + f"v1/people/{self.id}/features/voicemail/actions/uploadNoAnswerGreeting/invoke",
                          headers={"Content-Type": encoder.content_type, **self._headers},
                          data=encoder)
        content.close()
        if r.ok:
            if activate is True:
                vm_config = self.get_vm_config()
                vm_config['sendUnansweredCalls']['greeting'] = "CUSTOM"
                self.push_vm_config(vm_config)
            return True
        else:
            log.warning(f"The Greeting upload failed: {r.text}")
            return False

    def push_cf_config(self, cf_config: dict = None):
        """ Pushes the Call Forwarding config back to Webex

        If the cf_config dict is provided, it will be sent as the PUT payload. If it is omitted, the current
        Person.call_forwarding attributes will be sent

        Args:
            cf_config (dict, optional): The cf_config dictionary to push to Webex

        Returns:
            dict: The new config that was sent back by Webex

        """
        log.info(f"Pushing CF Config for {self.email}")
        if cf_config is not None:
            payload = cf_config
        else:
            payload = self.call_forwarding
        success = self.__put_webex_data(f"v1/people/{self.id}/features/callForwarding", payload)
        if success:
            self.get_call_forwarding()
            return self.call_forwarding
        else:
            return False

    def enable_vm_to_email(self, email: str = None, push=True):
        """
        Change the Voicemail config to enable sending a copy of VMs to specified email address. If the email param
        is not present, it will use the Person's email address as the default.

        Args:
            email (optional): The email address to send VMs to.
            push (optional): Whether to immediately push the change to Webex. Defaults to True.

        Returns:
            dict: The `Person.vm_config` attribute

        """
        if not self.vm_config:
            self.get_vm_config()
        if email is None:
            email = self.email
        self.vm_config['emailCopyOfMessage']['enabled'] = True
        self.vm_config['emailCopyOfMessage']['emailId'] = email
        if push:
            return self.push_vm_config()
        else:
            return self.vm_config

    def disable_vm_to_email(self, push=True):
        """ Change the Voicemail config to disable sending a copy of VMs to specified email address.

        Args:
            push (bool, optional): Whether to immediately push the change to Webex. Defaults to True.

        Returns:
            dict: The `Person.vm_config` attribute

        """
        log.info(f"Disabling VM-to-email for {self.email}")
        if not self.vm_config:
            self.get_vm_config()
        self.vm_config['emailCopyOfMessage']['enabled'] = False
        if push:
            return self.push_vm_config()
        else:
            return self.vm_config

    def enable_vm_notification(self, email: str = None, push=True):
        """
        Change the Voicemail config to enable voicemail notification to specified email address. If the email param
        is not present, it will use the Person's email address as the default.

        Args:
            email (optional): The email address to send VMs to.
            push (optional): Whether to immediately push the change to Webex. Defaults to True.

        Returns:
            dict: The `Person.vm_config` attribute
        """
        if not self.vm_config:
            self.get_vm_config()
        if email is None:
            email = self.email
        self.vm_config['notifications']['enabled'] = True
        self.vm_config['notifications']['destination'] = email
        if push:
            return self.push_vm_config()
        else:
            return self.vm_config

    def disable_vm_notification(self, email: str = None, push=True):
        """
        Change the Voicemail config to disable voicemail notification to specified email address. If the email param
        is not present, it will use the Person's email address as the default.

        Args:
            email (optional): The email address to send VMs to.
            push (optional): Whether to immediately push the change to Webex. Defaults to True.

        Returns:
            dict: The `Person.vm_config` attribute
        """
        if not self.vm_config:
            self.get_vm_config()
        if email is None:
            email = self.email
        self.vm_config['notifications']['enabled'] = False
        if push:
            return self.push_vm_config()
        else:
            return self.vm_config

    def set_voicemail_rings(self, rings: int, push=True):
        """ Set the number of rings before an Unanswered call is sent to Voicemail

        Args:
            rings (int): The number of rings
            push (bool, optional): Whether to immediately push the change to Webex. Defaults to True.

        Returns:
            dict: The `Person.vm_config` attribute

        """
        if not self.vm_config:
            self.get_vm_config()
        self.vm_config['sendUnansweredCalls']['numberOfRings'] = rings
        if push:
            return self.push_vm_config()
        else:
            return self.vm_config

    def get_intercept(self):
        """Gets the Intercept config for the Person

        Returns:
            dict: The Intercept config for the Person instance

        """
        log.info("get_intercept() started")
        self.intercept = self.__get_webex_data(f"v1/people/{self.id}/features/intercept")
        return self.intercept

    def push_intercept(self, config: dict):
        """ Push the Intercept config to Webex

        Returns:
            bool: True on success, False otherwise

        """
        log.info(f"Pushing Intercept config for {self.email}")
        success = self.__put_webex_data(f"v1/people/{self.id}/features/intercept", payload=config)
        if success:
            return True
        else:
            log.warning("The Intercept config push failed")
            return False

    def get_ptt(self):
        """ Gets Push-to-Talk config for the Person

        Returns:
            dict: The Push-to-Talk config for the Person instance

        """
        log.info("get_ptt() started")
        self.ptt = self.__get_webex_data(f"v1/people/{self.id}/features/pushToTalk")
        return self.ptt

    def push_ptt(self, config: dict):
        """ Push the Push-to-Talk config to Webex

        Args:
            config: The configuration to push

        Returns:
            bool: True on success, False otherwise

        """
        log.info(f"Pushing PTT config for {self.email}")
        success = self.__put_webex_data(f"v1/people/{self.id}/features/pushToTalk", payload=config)
        if success:
            return True
        else:
            log.warning("The PTT config push failed")
            return False

    def get_call_recording(self):
        """The Call Recording config for the Person

        Returns:
            dict: The Call Recording config for the Person instance

        """
        log.info(f"Getting Call Recording config for {self.email}")
        self.call_recording = self.__get_webex_data(f"v1/people/{self.id}/features/callRecording")
        return self.call_recording

    def push_call_recording(self, config: dict):
        """ Push the Call Recording config to Webex

        Args:
            config (dict): The Call Recording config as defined by the Webex API specification

        Returns:
            bool: True on success, False otherwise

        """
        log.info(f"Pushing Call Recording config for {self.email}")
        # The payload from Webex has the Dubber Service Provider info, which isn't supported by the PUT
        # We have to delete those keys if they exist before sending it
        clean_config = config.copy()
        clean_config.pop("serviceProvider", None)
        clean_config.pop("externalGroup", None)
        clean_config.pop("externalIdentifier", None)
        success = self.__put_webex_data(f"v1/people/{self.id}/features/callRecording", payload=clean_config)
        if success:
            return True
        else:
            log.warning("The Call Recording config push failed")
            return False

    @property
    def monitoring(self):
        """ :class:`~.monitoring.MonitoringList` to view and control monitoring """
        if self._monitoring is None:
            response = webex_api_call("get", f"v1/people/{self.id}/features/monitoring",
                                      params={"orgId": self.org_id})
            response['parent'] = self
            response['org'] = self._parent
            if "monitoredElements" not in response.keys():
                response['monitoredElements'] = []
            self._monitoring = MonitoringList.from_dict(response)
        return self._monitoring

    def get_monitored_by(self):
        """ Returns a list of Users (Person) and Workspaces that are Monitoring this Person """
        if isinstance(self._parent, wxcadm.Org):
            monitor_list = self._parent.get_all_monitoring()
        elif isinstance(self._parent, wxcadm.Location):
            monitor_list = self._parent.parent.get_all_monitoring()
        try:
            return monitor_list['people'][self.id]
        except KeyError:
            return None

    def get_hoteling(self) -> dict:
        """ Get the Hoteling config for the Person

        Returns:
            dict: The Hoteling config for the Person instance

        """
        log.info(f"Getting Hoteling config for {self.email}")
        self.hoteling = self.__get_webex_data(f"v1/people/{self.id}/features/hoteling")
        return self.hoteling

    def push_hoteling(self, config: dict) -> bool:
        """ Push the Hoteling config to Webex

        Args:
            config (dict): The Hoteling config as defined by the Webex API specification

        Returns:
            bool: True on success, False otherwise

        """
        log.info(f"Pushing Hoteling config for {self.email}")
        success = self.__put_webex_data(f"v1/people/{self.id}/features/hoteling", payload=config)
        if success:
            self.get_hoteling()
            return True
        else:
            log.warning("The Hoteling config push failed")
            return False

    def get_outgoing_permission(self):
        """ Get the Outgoing Calling Permission for the Person """
        log.info(f"Getting Outbound Calling Permission config for {self.email}")
        self._outgoing_permission = self.__get_webex_data(f"v1/people/{self.id}/features/outgoingPermission")
        return self._outgoing_permission

    def push_outgoing_permission(self, config: dict) -> bool:
        """ Sets the Outgoing Call Permission using the provided dict

        Args:
            config (dict): The Outgoing Call Permission config as defined by the Webex API specification

        Returns:
            bool: True on success, False otherwise.

        """
        log.info(f"Pushing Outgoing Calling Permission config for {self.email}")
        success = self.__put_webex_data(f"v1/people/{self.id}/features/outgoingPermission", payload=config)
        if success:
            self.get_outgoing_permission()
            return True
        else:
            log.warning("The Outgoing Permission push failed")
            return False

    def get_caller_id(self) -> dict:
        """ Get the Caller ID config for the Person

        Returns:
            dict: The Called ID config for the Person instance

        """
        log.info(f"Getting Caller ID config for {self.email}")
        self.caller_id = self.__get_webex_data(f"v1/people/{self.id}/features/callerId")
        return self.caller_id

    def push_caller_id(self, config: dict) -> bool:
        """ Push the Caller ID config to Webex

        This method differs from :meth:`Person.set_caller_id()`. It sends the config dict directly.
        :meth:`Person.set_caller_id()` builds the config dict itself so that you don't have to know all the
        intricacies of the API specification. In most cases, that method is preferable.

        Args:
            config (dict): The Caller ID config as defined by the Webex API specification

        Returns:
            bool: True on success, False otherwise

        """
        log.info(f"Pushing Caller Id config for {self.email}")
        success = self.__put_webex_data(f"v1/people/{self.id}/features/callerId", payload=config)
        if success:
            self.get_caller_id()
            return True
        else:
            log.warning("The Caller ID config push failed")
            return False

    def set_caller_id(self, name: str, number: str):
        """ Change the Caller ID for a Person

        Args:
            name (str): The name to set as the Caller ID Name. Also accepts keywords: ``direct`` sets the name to the
                user's name in Webex. ``location`` sets the name to the name of the Location.
            number (str): The number to set as the Caller ID.  Also accepts keywords: ``direct`` sets the number to the
                user's DID in Webex. ``location`` sets the name to the main number of the Location.

        Returns:
            bool: True on success

        Raises:
            wxcadm.exceptions.APIError: Raised when there is a problem with the API call

        """
        log.info(f"Setting Caller ID for {self.email}")
        log.debug(f"\tNew Name: {name}\tNew Number: {number}")
        payload = {}
        # Handle the possible number values
        if number.lower() == "direct":
            payload['selected'] = "DIRECT_LINE"
        elif number.lower() == "location":
            payload['selected'] = "LOCATION_NUMBER"
        else:
            payload['selected'] = "CUSTOM"
            payload['customNumber'] = number
        # Then deal with possible name values
        if name.lower() == "direct":
            payload['externalCallerIdNamePolicy'] = "DIRECT_LINE"
        elif name.lower() == "location":
            payload['externalCallerIdNamePolicy'] = "LOCATION"
        else:
            payload['externalCallerIdNamePolicy'] = "OTHER"
            payload['customExternalCallerIdName'] = name

        success = self.push_caller_id(config=payload)
        if success:
            return True
        else:
            return False

    def get_dnd(self) -> dict:
        """ Get the Do Not Disturb (DND) config for the Person

        Returns:
            dict: The DND config for the Person instance

        """
        log.info(f"Getting DND config for {self.email}")
        self.dnd = self.__get_webex_data(f"v1/people/{self.id}/features/doNotDisturb")
        return self.dnd

    def push_dnd(self, config: dict) -> bool:
        """ Push the Do Not Disturb (DND) config to Webex

        Args:
            config (dict): The DND configuration to push

        Returns:
            bool: True on success, False otherwise

        """
        log.info(f"Pushing DND config for {self.email}")
        log.debug(f"\tConfig: {config}")
        success = self.__put_webex_data(f"v1/people/{self.id}/features/doNotDisturb", payload=config)
        if success:
            return True
        else:
            log.warning("The DND config push failed")
            return False

    def get_calling_behavior(self) -> dict:
        """ Get the Calling Behavior config for the Person

        Returns:
            dict: The Calling Behavior config for the Person instance

        """
        log.info(f"Getting Calling Behavior for {self.email}")
        self.calling_behavior = self.__get_webex_data(f"v1/people/{self.id}/features/callingBehavior")
        return self.calling_behavior

    def push_calling_behavior(self, config: dict) -> bool:
        """ Push the Calling Behavior config to Webex

        Args:
            config (dict): The Calling Behavior config

        Returns:
            bool: True on success, False otherwise

        """
        log.info(f"Pushing Calling Behavior for {self.email}")
        log.debug(f"\tConfig: {config}")
        success = self.__put_webex_data(f"v1/people/{self.id}/features/callingBehavior", payload=config)
        if success:
            return True
        else:
            log.warning("The Calling Behavior config push failed")
            return False

    def get_applications_settings(self) -> dict:
        """ Get the Application Services settings for the Person

        Returns:
            dict: The Application Services settings for the Person instance

        """
        log.info(f"Getting Applications Settings for {self.email}")
        self.applications_settings = self.__get_webex_data(f"v1/people/{self.id}/features/applications")
        return self.applications_settings

    def push_applications_settings(self, config: dict) -> bool:
        """ Push the Applications Services settings to Webex

        Args:
            config (dict): The Applications Services Settings config

        Returns:
            bool: True on success, False otherwise

        """
        log.info(f"Pushing Applications Settings for {self.email}")
        log.debug(f"\tConfig: {config}")
        success = self.__put_webex_data(f"v1/people/{self.id}/features/applications", payload=config)
        if success:
            return True
        else:
            log.warning("The Applications Settings config push failed")
            return False

    def license_details(self):
        """Get the full details for all licenses assigned to the Person

        Returns:
            list[dict]: List of the license dictionaries

        """
        log.info(f"Getting license details for {self.email}")
        license_list = []
        for license in self.licenses:
            for org_lic in self._parent.licenses:
                if license == org_lic['id']:
                    license_list.append(org_lic)
        return license_list

    def refresh_person(self, raw: bool = False):
        """
        Pull a fresh copy of the Person details from the Webex API and update the instance. Useful when changes
        are made outside the script or changes have been pushed and need to get updated info.

        Args:
            raw (bool, optional): Return the "raw" config from the as a dict. Useful when making changes to
                the user, because you have to send all values again.

        Returns:
            bool: True if successful, False if not

        """
        response = self.__get_webex_data(f"v1/people/{self.id}")
        if response:
            self.__process_api_data(response)
            if raw:
                return response
            else:
                return True
        else:
            return False

    def update_person(self,
                      email=None,
                      numbers=None,
                      extension=None,
                      location=None,
                      display_name=None,
                      first_name=None,
                      last_name=None,
                      roles=None,
                      licenses=None,
                      avatar: Optional[str] = None,
                      department: Optional[str] = None,
                      title: Optional[str] = None,
                      addresses: Optional[list] = None):
        """Update the Person in Webex.

        Pass only the arguments that you want to change. Other attributes will be populated
        with the existing values from the instance. *Note:* This allows changes directly to the instance attrs to
        be pushed to Webex. For example, changing :py:attr:`Person.extension` and then calling `update_person()` with no
        arguments will still push the extension change. This allows for a lot of flexibility if a method does not
        exist to change the desired value. It is also the method other methods use to push their changes to Webex.

        Args:
            email (str): The email address of the Person
            numbers (list): The list of number dicts ("type" and "value" keys)
            extension (str): The user's extension
            location (str): The Location ID for the user. Note that this can't actually be changed yet.
            display_name (str): The Display Name for the Person
            first_name (str): The Person's first name
            last_name (str): The Person's last name
            roles (list): List of Role IDs
            licenses (list): List of License IDs
            avatar (str): The URL of the Person's avatar
            department (str): The Person's department
            title (str): The Person's title
            addresses (list): A list of addresses, each defined as a dict

        Returns:
            bool: True if successful. False if not.

        """
        # Build the payload using the arguments and the instance attrs
        payload = {}
        if email is None:
            email = self.email
        payload['emails'] = [email]
        if numbers is None:
            numbers = self.numbers
        payload['phoneNumbers'] = numbers
        if extension is None:
            if self.extension:
                extension = self.extension
        payload['extension'] = extension
        if location is None:
            location = self.location
        payload['locationId'] = location
        if display_name is None:
            display_name = self.display_name
        payload['displayName'] = display_name
        if first_name is None:
            first_name = self.first_name
        payload['firstName'] = first_name
        if last_name is None:
            last_name = self.last_name
        payload['lastName'] = last_name
        if roles is None:
            roles = self.roles
        payload['roles'] = roles
        if licenses is None:
            licenses = self.licenses
        payload['licenses'] = licenses
        if avatar is None:
            avatar = self.avatar
        payload['avatar'] = avatar
        if department is None:
            department = self.department
        payload['department'] = department
        if title is None:
            title = self.title
        payload['title'] = title
        if addresses is None:
            addresses = self.addresses
        payload['addresses'] = addresses

        params = {"callingData": "true"}
        response = webex_api_call('put', f'v1/people/{self.id}', payload=payload, params=params)
        if response:
            self.refresh_person()
            return True
        else:
            return False

    def set_calling_only(self) -> bool:
        """
        Removes the Messaging and Meetings licenses, leaving only the Calling capability.

        Returns:
            bool: True on success, False otherwise

        """
        log.info(f"Setting {self.email} to Calling-Only")
        # First, iterate the existing licenses and remove the ones we don't want
        # Build a list that contains the values to match on to remove
        remove_matches = ["messaging",
                          "meeting",
                          "free"]
        new_licenses = []
        for license in self.licenses:
            log.debug(f"Checking license: {license}")
            lic_name = self._parent.get_license_name(license)
            log.debug(f"License Name: {lic_name}")
            if any(match in lic_name.lower() for match in remove_matches):
                if "screen share" in lic_name.lower():
                    log.debug(f"{lic_name} matches but is needed")
                    new_licenses.append(license)
                else:
                    log.debug(f"License should be removed")
                    continue
            else:
                log.debug(f"Keeping license")
                new_licenses.append(license)

        success = self.update_person(licenses=new_licenses)
        if success:
            return True
        else:
            log.warning("The Set Calling Only command failed")
            return False

    def change_phone_number(self, new_number: str, new_extension: str = None):
        """ Change a person's phone number and extension

        Args:
            new_number (str): The new phone number for the person
            new_extension (str, optional): The new extension, if changing. Omit to leave the same value.

        Returns:
            bool: True on success, False otherwise

        """
        if not new_extension:
            if self.extension:
                extension = self.extension
            else:
                extension = None
        else:
            extension = new_extension
        log.info(f"Changing phone number for {self.email} to {new_number} with extension: {str(new_extension)}")

        # Call the update_person() method
        success = self.update_person(numbers=[{"type": "work", "value": new_number}], extension=extension)
        if success:
            self.refresh_person()
            return True
        else:
            log.warning("Updating the phone number config failed")
            return False

    def disable_call_recording(self):
        """ Disables Call Recording for the Person

        This method will return True even if the Person did not have Call Recording enabled prior to calling
        the method.

        Returns:
            bool: True on success, False otherwise

        """
        log.info(f"Disabling Call Recording for {self.email}")
        recording_config = {'enabled': False}
        response = self.push_call_recording(recording_config)
        self.call_recording = response
        return True

    def enable_call_recording(self, type: str,
                              record_vm: bool = False,
                              announcement_enabled: bool = False,
                              reminder_tone: bool = False,
                              reminder_interval: int = 30,
                              can_play: bool = True,
                              can_download: bool = True,
                              can_delete: bool = True,
                              can_share: bool = True,
                              transcribe: bool = True,
                              ai_summary: bool = True):
        """ Enable and configure Call Recording for the Person

        .. note::

            Some parameters, such as ``'transcribe'`` and ``'can_play'`` only apply to call recording done by the Webex
            platform. If the recording is being done by another provider, these parameters will have no effect.

        Args:
            type (str): The type of Call Recording. Value must be 'always', 'never', 'always_with_pause' or 'on_demand'
            record_vm (bool, optional): Whether to record Voicemail. Defaults to False
            announcement_enabled (bool, optional): Whether to announce Call Recording. Defaults to False
            reminder_tone (bool, optional): Whether to play a reminder tone. Defaults to False
            reminder_interval (int, optional): Interval in seconds between reminders. Defaults to 30
            can_play (bool, optional): Whether the user can play recordings. Defaults to True.
            can_download (bool, optional): Whether the user can download recordings. Defaults to True.
            can_delete (bool, optional): Whether the user can delete recordings. Defaults to True.
            can_share (bool, optional): Whether the user can share recordings. Defaults to True.
            transcribe (bool, optional): Enable AI transcription of recordings. Defaults to True.
            ai_summary (bool, optional): Enable AI summary of recordings. Defaults to True.

        Returns:
            bool: True on success, False otherwise

        """
        type_map = {"always": "Always",
                    "never": "Never",
                    "always_with_pause": "Always with Pause/Resume",
                    "on_demand": "On Demand with User Initiated Start"}
        if type not in type_map.keys():
            raise ValueError("'type' must be 'always', 'never', 'always_with_pause' or 'on_demand'.")
        payload = {"enabled": True,
                   "record": type_map[type],
                   "recordVoicemailEnabled": record_vm,
                   "startStopAnnouncementEnabled": announcement_enabled,
                   "notification": {
                       "type": "Beep",
                       "enabled": reminder_tone,
                       "repeat": {
                           "enabled": reminder_tone,
                           "interval": reminder_interval
                       }
                   },
                   "callRecordingAccessSettings": {
                       "viewAndPlayRecordingsEnabled": can_play,
                       "downloadRecordingsEnabled": can_download,
                       "deleteRecordingsEnabled": can_delete,
                       "shareRecordingsEnabled": can_share
                   },
                   "postCallRecordingSettings": {
                       "transcriptEnabled": transcribe,
                       "summaryAndActionItemsEnabled": ai_summary
                   }
                   }
        success = self.push_call_recording(payload)
        if success:
            return True
        else:
            return False

    def get_executive_assistant(self):
        """ Get the Executive Assistant config for the Person

        Returns:
            dict: The Executive Assistant config for the Person instance

        """
        log.info(f"Getting Executive Assistant config for {self.email}")
        self.executive_assistant = self.__get_webex_data(f"v1/people/{self.id}/features/executiveAssistant")
        return self.executive_assistant

    def push_executive_assistant(self, config: dict) -> bool:
        """ Push the Executive Assistant config to Webex

        Args:
            config (dict): The Executive Assistant Config

        Returns:
             bool: True on success, False otherwise

        """
        log.info(f"Pushing Executive Assistant config for {self.email}")
        log.debug(f"\tConfig: {config}")
        success = self.__put_webex_data(f"v1/people/{self.id}/features/executiveAssistant", payload=config)
        if success:
            return True
        else:
            log.warning("The Executive Assistant config push failed")
            return False

    @property
    def wxc_numbers(self):
        """ All the phone numbers for the user from the Webex Calling platform.

        Unlike the :py:meth:`Person.numbers` property, which only reads the data from Webex Common Identity, this
        property pulls the data from the Webex Calling platform, so it includes the Primary number as well as any
        Alias numbers and doesn't include any Active Directory data.

        Returns:
            dict: Dict of numbers properties as defined at
                developer.webex.com/docs/api/v1/webex-calling-person-settings/get-a-list-of-phone-numbers-for-a-person

        """
        return webex_api_call('get', f'/v1/people/{self.id}/features/numbers')

    def remove_did(self) -> str:
        """ Remove the DID (phone number) from the Person

        Returns:
            str: The phone number that was removed. Returned to make it easier to add after a User Move, for example.

        """
        old_numbers = self.numbers
        new_numbers = []
        update_needed = False
        removed_number: Optional[str] = None
        for number in old_numbers:
            if number['type'].lower() != 'work':
                new_numbers.append(number)
            else:
                update_needed = True
                removed_number = number['value']
        if update_needed is True:
            self.update_person(numbers=new_numbers)
        return removed_number

    def add_did(self, phone_number: str, primary: Optional[bool] = True):
        """ Add a DID (phone number) to the Person

        Args:
            phone_number (str): The phone number to add
            primary (bool, optional): Whether the number will be the Primary phone number. Defauils to True.

        Returns:

        """
        numbers = self.numbers
        numbers.append({'type': 'work', 'value': phone_number, 'primary': primary})
        self.update_person(numbers=numbers)
        self.refresh_person()
        return self.numbers

    @property
    def ecbn(self) -> dict:
        """ The Emergency Callback Number details of the Person """
        response = webex_api_call('get', f'v1/telephony/config/people/{self.id}/emergencyCallbackNumber',
                                  params={'orgId': self.org_id})
        return response

    def ecbn_null_change(self) -> bool:
        """ Issue a "null change" on the user's ECBN, sending the same values that exist currently.

        This was added due to a back-end bug discovered that caused ECBN settings to be out of sync between Control
        Hub and Webex Calling. In most cases, this will not be used, but can be used in specialized scripts to correct
        this issue.

        """
        log.info(f"Performing null ECBN change for user: {self.email}")
        current_ecbn = self.ecbn
        payload = {
            'selected': current_ecbn['selected']
        }
        if current_ecbn['selected'] == 'LOCATION_MEMBER_NUMBER':
            payload['locationMemberId'] = current_ecbn['locationMemberInfo']['memberId']
        response = webex_api_call('put', f"v1/telephony/config/people/{self.id}/emergencyCallbackNumber",
                                  params={'orgId': self.org_id}, payload=payload)
        log.debug(response)
        return True

    def set_ecbn(self, value: Union[str, wxcadm.Person, wxcadm.Workspace, wxcadm.VirtualLine]):
        """ Set the ECBN of the Person

        Valid values are ``'direct'``, ``'location'``, or a :class:`Person`, :class:`Workspace`, or
        :class:`VirtualLine` to set the ECBN to one of those.

        Args:
            value (str, Person, Workspace, VirtualLine): The value to set the ECBN to

        Returns:
            bool: True on success

        """
        if isinstance(value, wxcadm.Person) or \
                isinstance(value, wxcadm.Workspace) or \
                isinstance(value, wxcadm.VirtualLine):
            payload = {
                'selected': 'LOCATION_MEMBER_NUMBER',
                'locationMemberId': value.id
            }
        elif value.lower() == 'direct' or value.lower() == 'direct_line':
            payload = {'selected': 'DIRECT_LINE'}
        elif value.lower() == 'location' or value.lower() == 'location_ecbn':
            payload = {'selected': 'LOCATION_ECBN'}
        else:
            raise ValueError('Unknown value')

        response = webex_api_call('put', f'v1/telephony/config/people/{self.id}/emergencyCallbackNumber',
                                  params={'orgId': self.org_id}, payload=payload)
        return response


class Me(Person):
    """ The class representing the token owner. Some methods are only available at an owner scope. """

    def __init__(self, user_id, parent: wxcadm.Org = None, config: dict = None):
        super().__init__(user_id, parent, config)

    def get_voice_messages(self, unread: bool = False):
        """ Get all the Voice Messages for the :py:class:`Me` instance

        Args:
            unread (bool, optional): Whether to only return Unread messages. Default is False

        Returns:
            list[VoiceMessage]: A list of all :py:class:`VoiceMessage` instances

        """
        messages = []
        # Something funky about this API call needing more headers
        data = webex_api_call("get", "v1/telephony/voiceMessages")
        for item in data:
            if item['read'] is True and unread is True:
                continue
            else:
                m = VoiceMessage(**item)
            messages.append(m)
        return messages

    @property
    def voicemail_summary(self) -> Optional[dict]:
        """ A summary of the number of read and unread Voice Messages

        Returns:
            dict: A dict of the Voice Message summary in the format::

                {
                    'newMessages': int,
                    'oldMessages': int,
                    'newUrgentMessages': int,
                    'oldUrgentMessages': int
                }

        """
        data: dict = webex_api_call('get', 'v1/telephony/voiceMessages/summary')
        if data:
            return data
        else:
            return None


@dataclass
class VoiceMessage:
    id: str
    """ The unique identifier for the Voice Message """
    duration: int
    """ The duration (in seconds) of the Voice Message. Not present for a fax message."""
    callingParty: dict
    """ The Calling Party's details """
    read: bool
    """ True is the Voice Message has been read/heard """
    created: str
    """ The date and time the Voice Message was created """
    urgent: bool = False
    """ True if the Voice Message has been marked Urgent """
    confidential: bool = False
    """ True if the Voice Message has been marked Confidential """
    faxPageCount: Optional[int] = None
    """ Number of pages in the fax. On"""

    def mark_read(self) -> bool:
        """ Mark the message as read within Webex

        Returns:
            bool: True on success, False otherwise

        """
        log.info(f"Marking Voice Message as read: {self.id}")
        payload = {"messageId": self.id}
        success = webex_api_call("post", "v1/telephony/voiceMessages/markAsRead", payload=payload)
        if success:
            return True
        else:
            log.warning("Something went wrong marking the message read.")
            return False

    def mark_unread(self) -> bool:
        """ Mark the message as unread within Webex

        Returns:
            bool: True on success, False otherwise

        """
        log.info(f"Marking Voice Message as unread: {self.id}")
        payload = {"messageId": self.id}
        success = webex_api_call("post", "v1/telephony/voiceMessages/markAsUnread", payload=payload)
        if success:
            return True
        else:
            log.warning("Something went wrong marking the message unread.")
            return False

    def delete(self) -> bool:
        """ Delete the Voice Message

        Returns:
            bool: True on success, False otherwise

        """
        log.info(f"Deleting Voice Message with ID {self.id}")
        success = webex_api_call("delete", f"v1/telephony/voiceMessages/{self.id}")
        if success:
            return True
        else:
            log.warning("The Voice Message delete failed")
            return False


class UserGroups(UserList):
    """ UserGroups is the parent class for :py:class:`UserGroup`, providing methods for the list of Groups """

    def __init__(self, parent: wxcadm.Org):
        log.info("Initializing UserGroups instance")
        super().__init__()
        self.parent = parent
        self.data: list[UserGroup] = []
        response = webex_api_call("get", "v1/groups", params={'orgId': self.parent.org_id})
        groups = response['groups']
        log.debug(f"Webex returned {len(groups)} Groups")
        for group in groups:
            usergroup = UserGroup(parent=self.parent, **group)
            self.data.append(usergroup)

    def create_group(self, name: str,
                     description: str = '',
                     members: Optional[list] = None) -> bool:
        """ Create a new UserGroup

        Args:
            name (str): The name of the Group
            description (str, optional): An optional description of the Group
            members (list[Person], optional): An optional list of :py:class:`Person` instances to add to the Group

        Returns:
            bool: True on success, False otherwise

        """
        payload = {'displayName': name, 'orgId': self.parent.id, 'description': description}
        if members is not None:
            payload['members'] = []
            for member in members:
                payload['members'].append({'id': member.id})
        response = webex_api_call('post', 'v1/groups', payload=payload)
        if response:
            log.info(f"New UserGroup {name} created")
            new_group = UserGroup(parent=self.parent, **response)
            self.data.append(new_group)
            return True
        else:
            log.info("Failed to create new UserGroup")
            return False

    def find_person_assignments(self, person: Person):
        assignments = []
        for group in self.data:
            for member in group.members:
                if isinstance(member, wxcadm.Person):
                    if member.id == person.id:
                        assignments.append(group)
        return assignments


@dataclass()
class UserGroup:
    """ The UserGroup class holds all the User Groups available within the Org"""
    parent: wxcadm.Org = field(repr=False)
    """ The Org instance that owns the Group """
    id: str
    """ The unique ID of the Group """
    displayName: str
    """ The name of the Group """
    orgId: str = field(repr=False)
    """ The Org ID to which the Group belongs """
    created: str = field(repr=False)
    """ The timestamp indicating when the Group was created """
    lastModified: str = field(repr=False)
    """ The timestamp indicating the last time the Group was modified """
    # Changed 3.0.0 - This field no longer comes back in the API response. Not sure if it will be re-added
    usage: str = field(init=True, repr=True, default='Unknown')
    """ The Group usage type. This is a value that was provided by Webex but has been removed """
    memberSize: int = field(init=True, repr=False, default=0)
    """ The number of members in the group only if returned by Webex """
    description: str = field(repr=False, default='')
    """ The long description of the Group """

    @property
    def members(self):
        """ List of all members with this Group """
        return self._get_members()

    @property
    def name(self):
        """ The name of the UserGroup """
        return self.displayName

    def _get_members(self):
        response = webex_api_call("get", f"/v1/groups/{self.id}/members")
        items = response['members']
        # If there are more than 500 members, we need to go get the rest, and the Groups API handles this
        # differently than all the other APIs. This is probably going to break things if they ever fix the Groups
        # API, but we'll handle that when it happens

        people = []
        for item in items:
            person = self.parent.people.get(id=item['id'])
            if person is not None:
                people.append(person)
            else:
                # Leaving the following here in case Groups ever support Workspaces
                # workspace = self.parent.workspaces.get(id=item['id'])
                # if workspace is not None:
                #     people.append(workspace)
                # else:
                people.append(item['id'])
        return people

    def delete(self) -> bool:
        """ Delete the Group

        Returns:
            bool: True on success, False otherwise

        """
        response = webex_api_call('delete', f'v1/groups/{self.id}')
        if response:
            log.info(f'Successfully deleted UserGroup {self.displayName}')
            return True
        else:
            log.warning(f'Failed to delete UserGroup {self.displayName}')
            return False

    def add_member(self, person: Person) -> bool:
        """ Add a Person to the Group

        Args:
            person (Person): The :py:class:`Person` instance to add

        Returns:
            bool: True on success

        Raises:
            ValueError: Raised when trying to add a Person to a Location Group

        """
        payload = {"members": [{"id": person.id, "operation": "add"}]}
        response = webex_api_call("patch", f"v1/groups/{self.id}", payload=payload)
        if response:
            return True
        else:
            return False

    def delete_member(self, person: Person) -> bool:
        """ Delete the specified Person from the Group

        Args:
            person (Person): The :py:class:`Person` instance to delete

        Returns:
            bool: True on success, False otherwise

        """
        payload = {"members": [{'id': person.id, 'operation': 'delete'}]}
        response = webex_api_call('patch', f'v1/groups/{self.id}', payload=payload)
        if response:
            return True
        else:
            return False



from __future__ import annotations

import requests
import wxcadm
from .exceptions import *
from .common import *


class Bifrost:
    """The Bifrost class handles API calls using the Bifrost API."""

    def __init__(self, org: wxcadm.Org, access_token: str):
        self._parent = org
        self._access_token = access_token
        self._headers = {"Authorization": f"Bearer {access_token}"}

        # Calculate the "customer" ID from the Org ID
        spark_id = decode_spark_id(org.id)
        self._customer = spark_id.split("/")[-1]

        self._url_base = f"https://bifrost-a.wbx2.com/api/v2/customers/{self._customer}/"
        self._server = "https://bifrost-a.wbx2.com"

    def get_location(self):
        locations = []
        params = {"limit": 2000, "offset": 0}   # Default values for the numbers pull

        get_more = True     # Bool to let us know to keep pulling more numbers
        next_url = None
        while get_more:
            if next_url is None:
                r = requests.get(self._url_base + f"locations", headers=self._headers, params=params)
            else:
                r = requests.get(next_url, headers=self._headers)
            if r.status_code == 200:
                response = r.json()
                locations.extend(response['locations'])
                if "next" in response['paging']:
                    get_more = True
                    next_url = response['paging']['next']
                else:
                    get_more = False
            elif r.status_code == 403:
                raise TokenError("Your API Access Token doesn't have permission to use this API call")
            else:
                raise APIError(f"The Bifrost locations call did not return a successful value")

        for location in locations:
            if "name" in location:
                webex_location = self._parent.locations.get(name=location['name'])
                webex_location.bifrost_config = location
        return locations



from __future__ import annotations

from typing import Optional, TYPE_CHECKING, Union

if TYPE_CHECKING:
    from .number import NumberList
from collections import UserList
from dataclasses import dataclass, field
from dataclasses_json import dataclass_json, config

import wxcadm
from .common import *
from wxcadm import log


@dataclass_json
@dataclass
class PSTNProvider:
    id: str
    """ The ID of the PSTN Provider """
    name: str = field(metadata=config(field_name="displayName"))
    """ The name of the PSTN Provider"""
    services: list = field(metadata=config(field_name="pstnServices"), default_factory=list)
    """ The services offered by the PSTN Provider """


class PSTNProviderList(UserList):
    def __init__(self, location: wxcadm.Location):
        super().__init__()
        self.parent = location
        self.data = []
        response = webex_api_call('get', f'v1/telephony/pstn/locations/{self.parent.id}/connectionOptions',
                                  params={'orgId': self.parent.org_id})
        for entry in response:
            self.data.append(PSTNProvider.from_dict(entry))

    def get(self, id: Optional[str] = None, name: Optional[str] = None) -> Optional[PSTNProvider]:
        """ Get a PSTN Provider by ID or Name

        Args:
            id (str, optional): The ID of the PSTN Provider
            name (str, optional): The Name of the PSTN Provider

        Returns:
            PSTNProvider: The PSTN Provider

        """
        if id is not None:
            for entry in self.data:
                if entry.id == id:
                    return entry
            return None
        if name is not None:
            for entry in self.data:
                if entry.name == name:
                    return entry
            return None
        return None


class LocationPSTN:
    def __init__(self, location: wxcadm.Location):
        self.parent = location
        self.available_providers = PSTNProviderList(location)
        """ The :class:`PSTNProviderList` of available providers for the Location """
        log.info(f"Getting PSTN information for Location: {self.parent.name}")
        try:
            response = webex_api_call('get', f'v1/telephony/pstn/locations/{self.parent.id}/connection',
                                      params={'orgId': self.parent.org_id})
            log.debug(f"Response: {response}")
            self.provider = PSTNProvider.from_dict(response)
            self.type = response['pstnConnectionType']
        except wxcadm.APIError as e:
            if "PSTN for the location is not configured" in e.args[0].get('errorMessage', ''):
                log.info("No PSTN configured for Location")
                self.provider = None
            else:
                log.warning(f"API Error: {e.args}")
                raise e

    def set_provider(self, provider: Union[PSTNProvider, str, wxcadm.RouteGroup, wxcadm.Trunk]):
        """ Set the PSTN Provider for the Location

        The Provider can be expressed as a string, which is the Provider's ID, a :class:`PSTNProvider`, a
        :class:`~.routing.Trunk` or a :class:`~.routing.RouteGroup`.

        Args:
            provider (str, PSTNProvider, RouteGroup, Trunk): The new Provider, Route Group or Trunk to use

        Returns:
            bool: True on success, False otherwise

        """
        log.info(f"Setting new PSTN Provider for Location {self.parent.name}")
        if isinstance(provider, str):
            provider = self.available_providers.get(id=provider)
            if provider is None:
                raise ValueError(f"Provider with ID {provider} not found")
        if isinstance(provider, wxcadm.Trunk) or isinstance(provider, wxcadm.RouteGroup):
            if isinstance(provider, wxcadm.Trunk):
                payload = {
                    'premiseRouteType': 'TRUNK'
                }
            else:
                payload = {
                    'premiseRouteType': 'ROUTE_GROUP'
                }
            payload['premiseRouteId'] = provider.id
        else:
            payload = {
                'id': provider.id,
            }
        webex_api_call('put', f"v1/telephony/pstn/locations/{self.parent.id}/connection",
                       payload=payload,
                       params={'orgId': self.parent.org_id})
        self.provider = provider
        return True





from __future__ import annotations

from typing import Optional, TYPE_CHECKING

if TYPE_CHECKING:
    from .number import NumberList
from collections import UserList
from types import SimpleNamespace
from typing import Union

import wxcadm
from wxcadm import log
from .models import LocationEmergencySettings
from .location_features import LocationSchedule, CallParkExtension, VoicePortal, OutgoingPermissionDigitPatternList
from .call_queue import CallQueueList
from .hunt_group import HuntGroupList
from .auto_attendant import AutoAttendantList
from .pickup_group import PickupGroupList
from .common import *
from .exceptions import APIError
from .dect import DECTNetworkList
from .number import NumberList
from .virtual_line import VirtualLineList
from .call_routing import TranslationPatternList
from .workspace import WorkspaceList
from .recording import LocationRecordingVendorSelection
from .pstn import LocationPSTN
from .number import Number


class LocationList(UserList):
    def __init__(self, parent: wxcadm.Org):
        super().__init__()
        log.debug("Initializing LocationList instance")
        self.parent: wxcadm.Org = parent
        self.data: list = self._get_items()

    def refresh(self):
        """ Refresh the list of Locations from Webex """
        self.data = self._get_items()

    def _get_items(self):
        if isinstance(self.parent, wxcadm.Org):
            log.debug("Using Org ID as data filter")
            params = {'orgId': self.parent.id}
        else:
            raise ValueError("Unsupported parent class")

        log.debug("Getting Location list")
        response = webex_api_call('get', f'v1/locations', params=params)
        log.debug(f"Received {len(response)} entries")
        items = []
        for entry in response:
            items.append(Location(parent=self.parent,
                                  location_id=entry['id'],
                                  name=entry['name'],
                                  address=entry['address'],
                                  time_zone=entry.get('timeZone', 'Unknown'),
                                  preferred_language=entry.get('preferredLanguage', 'en_US')))
        return items

    def get(self, id: str = None, name: str = None, spark_id: str = None):
        """ Get the Location instance associated with a given ID, Name, or Spark ID

        Only one parameter should be supplied in normal cases. If multiple arguments are provided, the Locations will be
        searched in order by ID, Name, and finally Spark ID. If no arguments are provided, the method will raise an
        Exception.

        Args:
            id (str, optional): The Location ID to find
            name (str, optional): The Location Name to find
            spark_id (str, optional): The Spark ID to find

        Returns:
            Location: The Location instance correlating to the given search argument.

        Raises:
            ValueError: Raised when the method is called with no arguments

        """
        if id is None and name is None and spark_id is None:
            raise ValueError("A search argument must be provided")
        if id is not None:
            for location in self.data:
                if location.id == id:
                    return location
        if name is not None:
            for location in self.data:
                if location.name == name:
                    return location
        if spark_id is not None:
            for location in self.data:
                if location.spark_id == spark_id:
                    return location
        return None

    def create(self,
               name: str,
               time_zone: str,
               preferred_language: str,
               address: dict,
               latitude: Optional[str] = None,
               logitude: Optional[str] = None):
        """ Create a new Location

        Args:
            name (str): The name of the Location
            time_zone (str): The Time Zone of the Location (e.g. America/Chicago)
            preferred_language (str): The preferred language for phones and menus (i.e. en_US)
            address (dict): A dictionary containing the address elements. The format of this dict should be::

                {
                    "address1": "100 N. Main",
                    "address2": "Suite 200",
                    "city": "Houston",
                    "state": "TX",
                    "postalCode": "32123",
                    "country": "US"
                }

            latitude (str, optional): The latitude of the location. If not specified, the default value is None.
            longitude (str, optional): The longitude of the location. If not specified, the default value is None.

        Returns:
            dict: The response from Webex with the Location details.

        """
        payload = {
            'name': name,
            'timeZone': time_zone,
            'preferredLanguage': preferred_language,
            'announcementLanguage': None,
            'address': address
        }
        response = webex_api_call('post', 'v1/locations', payload=payload)
        self.refresh()
        return self.get(id=response['id'])

    def webex_calling(self, enabled: bool = True, single: bool = False) -> Location | list[Location]:
        """ Return a list of :py:class:`Location` where Webex Calling is enabled/disabled

        Args:
            enabled (bool, optional): True (default) returns Webex Calling Locations. False returns Locations without
                Webex Calling

            single (bool, optional): When True, returns only a single Location, which can be useful for some API calls,
                such as the Device Password API.

        Returns:
            list[:py:class:`Location`]: List of :py:class:`Location` instances. An empty list is returned if none match
            the given criteria

            :py:class:`Location`: When ``single=True`` is present, a single Location will be returned.

        """
        # The following API call was added in 4.3.9 because the previous method required an API call for every Location
        # which was very slow. The new API call gets the Webex Calling config for all Locations, so it is assumed
        # that any Location returned has Webex Calling
        response = wxcadm.webex_api_call('get', 'v1/telephony/config/locations', params={'orgId': self.parent.org_id})
        locations = []
        entry: dict
        for entry in response['locations']:
            location = self.get(id=entry['id'])
            if location is not None:
                location.calling_enabled = True
                locations.append(location)
        # Return the Locations we just found if asked for it
        if enabled is True:
            if single is False:
                return locations
            elif single is True:
                return locations[0]
        elif enabled is False:
            locations = []
            loc: Location
            for loc in self.data:
                if loc.calling_enabled is False or loc.calling_enabled is None:
                    loc.calling_enabled = False
                    locations.append(loc)
            if single is False:
                return locations
            elif single is True:
                return locations[0]

    def with_pstn(self, has_pstn: bool = True):
        """ Return o list of :class:`Location` where the location has PSTN enabled

        Args:
            has_pstn (bool): True for Locations with PSTN, False for Locations without PSTN. Defaults to True.

        Returns:
            list[:class:`Location`]: List of :class:`Location` instances. An empty list is returned if none match

        """
        with_pstn = []
        for loc in self.webex_calling(enabled=True):
            if has_pstn is True:
                if loc.pstn.provider is not None:
                    with_pstn.append(loc)
            if has_pstn is False:
                if loc.pstn.provider is None:
                    with_pstn.append(loc)
        return with_pstn

    def without_pstn(self):
        """ Alias function of :meth:`LocationList.with_pstn(False)` to quickly provide Locations without PSTN """
        return self.with_pstn(has_pstn=False)

class Location:
    def __init__(self, parent: wxcadm.Org, location_id: str,
                 name: str,
                 time_zone: str,
                 preferred_language: str,
                 announcement_language: str = None,
                 address: dict = None):
        """Initialize a Location instance

        .. note::
            Location instances are normally not instantiated manually and are done automatically with the
            :meth:`Org.get_locations` method.

        Args:
            location_id (str): The Webex ID of the Location
            name (str): The name of the Location
            time_zone (str): The time zone of the Location
            preferred_language (str): The preferred language at the Location
            announcement_language (str): The language for audio announcements at the Location
            address (dict): The address information for the Location

        Returns:
             Location (object): The Location instance

        """
        self._parent = parent
        self.parent = parent
        self._headers = parent._headers
        self.id: str = location_id
        """The Webex ID of the Location"""
        self.name: str = name
        """The name of the Location"""
        self.address: dict = address
        """The address of the Location"""
        self.time_zone: str = time_zone
        """ The Location time zone"""
        self.preferred_language: str = preferred_language
        """ The preferred language at the Location"""
        self.announcement_language: str = announcement_language
        """ The language for audio announcements at the Location"""
        self._calling_enabled: Optional[bool] = None
        self._calling_config: Optional[dict] = None
        self._pickup_groups: Optional[PickupGroupList] = None
        self._call_queues: Optional[CallQueueList] = None
        self._hunt_groups: Optional[HuntGroupList] = None
        self._dect_networks: Optional[DECTNetworkList] = None
        self._outgoing_permission_digit_patterns: Optional[OutgoingPermissionDigitPatternList] = None
        self._numbers: Optional[NumberList] = None
        self._virtual_lines: Optional[VirtualLineList] = None
        self._floors: Optional[LocationFloorList] = None
        self._translation_patterns = None
        self._workspaces = None
        self._routing_prefix = None
        self._pstn = None
        self._main_number = None

    def __str__(self):
        return self.name

    def __repr__(self):
        return self.id

    @property
    def outgoing_permission_digit_patterns(self) -> OutgoingPermissionDigitPatternList:
        if self._outgoing_permission_digit_patterns is None:
            self._outgoing_permission_digit_patterns = OutgoingPermissionDigitPatternList(self)
        return self._outgoing_permission_digit_patterns

    @property
    def workspace_location(self):
        """ Get the :class:`WorkspaceLocation` associated with this Location

        Returns:
            WorkspaceLocation: The WorkspaceLocation associated with this Location

        """
        return self.parent.workspace_locations.get(name=self.name)

    @property
    def virtual_lines(self) -> VirtualLineList:
        """ The :class:`VirtualLineList` of Virtual Lines for this Location

        Returns:
            VirtualLineList: The list of Virtual Lines
        """
        if self._virtual_lines is None:
            self._virtual_lines = VirtualLineList(self)
        return self._virtual_lines

    @property
    def floors(self) -> LocationFloorList:
        """ The :class:`LocationFloorList` of floors for this Location """
        if self._floors is None:
            self._floors = LocationFloorList(self)
        return self._floors

    @property
    def people(self) -> list:
        return self.parent.people.get(location=self)

    @property
    def calling_enabled(self) -> bool:
        """ Whether the Location is enabled for Webex Calling """
        if self._calling_enabled is None:
            try:
                response = webex_api_call("get", f"v1/telephony/config/locations/{self.id}",
                                          params={'orgId': self.org_id})
                if 'id' in response.keys():
                    self._calling_enabled = True
                    self._calling_config = response
            except APIError:
                self._calling_enabled = False
                self._calling_config = None
        return self._calling_enabled

    @calling_enabled.setter
    def calling_enabled(self, value: bool):
        self._calling_enabled = value
        return

    @property
    def ecbn(self):
        """ The Emergency Callback Number details for the Location """
        response = webex_api_call('get', f"v1/telephony/config/locations/{self.id}/features/emergencyCallbackNumber",
                                  params={'orgId': self.org_id})
        return response

    def set_ecbn(self, value: Union[str, wxcadm.Person, wxcadm.Workspace, wxcadm.VirtualLine]) -> bool:
        """ Set the ECBN of the Location

        Valid values are ``'location'``, or a :class:`Person`, :class:`Workspace`, or :class:`VirtualLine` to set the
        ECBN to one of those.

        Args:
            value (str, Person, Workspace, VirtualLine): The value to set the ECBN to

        Returns:
            bool: True on success

        """
        if isinstance(value, wxcadm.Person) or \
                isinstance(value, wxcadm.Workspace) or \
                isinstance(value, wxcadm.VirtualLine):
            payload = {
                'selected': 'LOCATION_MEMBER_NUMBER',
                'locationMemberId': value.id
            }
        elif value.lower() == 'location' or value.lower() == 'location_number':
            payload = {'selected': 'LOCATION_NUMBER'}
        else:
            raise ValueError('Unknown value')

        response = webex_api_call('put', f'v1/telephony/config/locations/{self.id}/features/emergencyCallbackNumber',
                                  params={'orgId': self.org_id}, payload=payload)
        return True

    @property
    def unknown_extension_policy(self) -> dict:
        """ The Location configuration for routing unknown extensions to the Premises as internal calls

        The dict contains two keys, ``'enabled'`` and ``'route'``. The ``'route'`` key will be either a
        :class:`Trunk` or a :class:`RouteGroup` when the ``'enabled'`` values is ``True``.

        """
        policy = {}
        response = webex_api_call('get', f"v1/telephony/config/locations/{self.id}/internalDialing",
                                  params={'orgId': self.org_id})
        policy['enabled'] = response.get('enableUnknownExtensionRoutePolicy', False)
        policy['route'] = None
        if 'unknownExtensionRouteIdentity' in response.keys():
            identity = response['unknownExtensionRouteIdentity']
            entity_type = identity.get('type', '')
            if entity_type == 'ROUTE_GROUP':
                entity = self.parent.call_routing.route_groups.get(id=identity['id'])
            elif entity_type == 'TRUNK':
                entity = self.parent.call_routing.trunks.get(id=identity['id'])
            else:
                entity = 'UNKNOWN'
            policy['route'] = entity
        return policy

    def set_unknown_extension_policy(self, route: Union[wxcadm.Trunk, wxcadm.RouteGroup, str]) -> bool:
        """ Set the Unknown Extension Policy for the Location

        If the ``route`` parameter is a :class:`Trunk` or a :class:`RouteGroup`, the routing will be enabled and the
        provided :class:`Trunk` or :class:`RouteList` will be used. To disable Unknown Extension routing, pass the
        ``route`` argument a value of ``'disabled'``.

        Args:
            route: A :class:`Trunk`, :class:`RouteGroup` to use for routing, or ``'disabled'`` to disable routing

        Returns:
            bool: True on success

        """
        if isinstance(route, str):
            if route.lower() != 'disabled':
                raise ValueError("Invalid route type")
            else:
                payload = {'enableUnknownExtensionRoutePolicy': False}
        else:
            if isinstance(route, wxcadm.Trunk):
                route_type = 'TRUNK'
            elif isinstance(route, wxcadm.RouteGroup):
                route_type = 'ROUTE_GROUP'
            else:
                raise ValueError("Invalid route type")
            payload = {
                'enableUnknownExtensionRoutePolicy': True,
                'unknownExtensionRouteIdentity': {
                    'id': route.id,
                    'type': route_type
                }
            }
        webex_api_call('put', f"v1/telephony/config/locations/{self.id}/internalDialing",
                       params={'orgId': self.org_id}, payload=payload)
        return True

    @property
    def calling_config(self) -> dict:
        """ The Webex Calling configuration dict """
        if self._calling_config is None:
            try:
                response = webex_api_call("get", f"v1/telephony/config/locations/{self.id}",
                                          params={'orgId': self.org_id})
                if 'id' in response.keys():
                    self._calling_enabled = True
                    self._calling_config = response
            except APIError:
                self._calling_enabled = False
                self._calling_config = None
        return self._calling_config

    @property
    def external_caller_id_name(self) -> str:
        """ The name sent as Caller ID for External calls that do not have a value set """
        return self.calling_config.get('externalCallerIdName', None)

    def set_external_caller_id_name(self, name: str) -> bool:
        """ Set the name to be sent as Caller ID for External calls that do not have a value set

        Args:
            name: The name to send

        Returns:
            bool: True on success

        """
        payload = {'externalCallerIdName': name}
        webex_api_call('put', f"v1/telephony/config/locations/{self.id}",
                       payload=payload, params={'orgId': self.org_id})
        self._calling_config = None
        return True

    @property
    def routing_prefix(self) -> Optional[str]:
        """ The Routing Prefix (i.e. Location Code) for the Location """
        if self._routing_prefix is None:
            self._routing_prefix = self.calling_config.get('routingPrefix', None)
        return self._routing_prefix

    def set_routing_prefix(self, prefix: str) -> bool:
        """ Set the Routing Prefix for the Location

        Args:
            prefix (str): The new Routing Prefix

        Returns:
            bool: True on success

        """
        payload = {'routingPrefix': prefix}
        webex_api_call('put', f"v1/telephony/config/locations/{self.id}",
                       payload=payload, params={'orgId': self.org_id})
        self._routing_prefix = prefix
        return True

    @property
    def org_id(self):
        """ The Org ID to which the Location belongs. :attr:`parent.id` can also be used. """
        return self.parent.id

    def enable_webex_calling(self) -> bool:
        """ Enable the Location for Webex Calling

        If the Location is already a Webex Calling Location, no action will be taken and the method will return True.

        Returns:
            bool: True on success

        """
        if self.calling_enabled is False:
            log.info(f"Enabling Webex Calling for Location {self.name}")
            # Clean up the address. Found a failure when address2 fails when it is ''
            if self.address['address2'] == '':
                self.address['address2'] = None
            payload = {
                "id": self.id,
                "name": self.name,
                "timeZone": self.time_zone,
                "preferredLanguage": self.preferred_language.lower(),
                "announcementLanguage": self.preferred_language.lower(),
                "address": self.address
            }
            webex_api_call("post", "v1/telephony/config/locations", payload=payload, params={'orgId': self.org_id})
            self._calling_enabled = True
        return True

    def get_all_monitoring(self) -> dict:
        """ Returns a dict of all Users and Workspaces that are being monitored. The User (Person) or Workspace is the
        dict key and the Users and Workspaces that are monitoring that key are a list.

        Returns:
            dict: A dict in the format ``{ 'people': { person: [] }, 'workspaces': { person: [] } }``

        """
        all_monitoring = {'people': {}, 'workspaces': {}}
        for person in self.people:
            monitoring = person.get_monitoring()
            for monitor in monitoring.get('monitoredElements', []):
                if 'member' in monitor.keys():
                    if monitor['member']['type'] == 'PEOPLE':
                        target_person = self.parent.people.get(id=monitor['member']['id'])
                        if target_person not in all_monitoring['people'].keys():
                            all_monitoring['people'][target_person] = []
                        all_monitoring['people'][target_person].append(person)
                    elif monitor['member']['type'] == 'PLACE':
                        target_workspace = self.parent.workspaces.get(id=monitor['member']['id'])
                        if target_workspace not in all_monitoring['workspaces'].keys():
                            all_monitoring['workspaces'][target_workspace] = []
                        all_monitoring['workspaces'][target_workspace].append(person)
        for workspace in self.workspaces.webex_calling():
            monitoring = workspace.monitoring
            for monitor in monitoring.get('monitoredElements', []):
                if 'member' in monitor.keys():
                    if monitor['member']['type'] == 'PEOPLE':
                        target_person = self.parent.people.get(id=monitor['member']['id'])
                        if target_person not in all_monitoring['people'].keys():
                            all_monitoring['people'][target_person] = []
                        all_monitoring['people'][target_person].append(workspace)
                    elif monitor['member']['type'] == 'PLACE':
                        target_workspace = self.parent.workspaces.get(id=monitor['member']['id'])
                        if target_workspace not in all_monitoring['workspaces'].keys():
                            all_monitoring['workspaces'][target_workspace] = []
                        all_monitoring['workspaces'][target_workspace].append(workspace)
        return all_monitoring

    def delete(self):
        """ Delete a Location

        .. warning::

            There is currently no way to delete a Location outside of Control Hub. The method is defined here so users
            aren't looking for it expecting it to be there. It will always return False. When the API is exposed,
            this method will be updated to support it.

        Returns:
            bool: Always returns False and no action is taken on Webex

        """
        return False

    @property
    def spark_id(self):
        """The ID used by all underlying services."""
        return decode_spark_id(self.id)

    @property
    def hunt_groups(self):
        """ :class:`HuntGroupList` list of :class:`HuntGroup` instances for this Location """
        log.info(f"Getting Hunt Groups for Location: {self.name}")
        if self.calling_enabled is False:
            log.debug("Not a Webex Calling Location")
            return None
        self._hunt_groups = HuntGroupList(self)
        return self._hunt_groups

    @property
    def announcements(self):
        """ List of :py:class:`Announcement` instances for the Location

        .. note::
            This does not include announcements built at the Organization level

        """
        annc_list = self._parent.announcements.get_by_location_id(self.id)
        return annc_list

    @property
    def translation_patterns(self):
        """ The :class:`TranslationPatternList` with all Translation Patterns for the Location """
        if self._translation_patterns is None:
            self._translation_patterns = TranslationPatternList(self)
        return self._translation_patterns

    @property
    def auto_attendants(self):
        """ List of AutoAttendant instances for this Location """
        log.info(f"Getting Auto Attendants for Location: {self.name}")
        return AutoAttendantList(self)

    @property
    def call_queues(self) -> Optional[CallQueueList]:
        """ :class:`CallQueueList` of :class:`CallQueue` instances for this Location """
        if self._call_queues is None:
            log.info(f"Getting Call Queues for Location: {self.name}")
            if self.calling_enabled is False:
                log.warn("Not a Webex Calling Location")
                return None
            self._call_queues = CallQueueList(self)
        return self._call_queues

    @property
    def numbers(self) -> Optional[NumberList]:
        """ All the Numbers for the Location

        Returns:
            NumberList: The :class:`NumberList` for the Location

        """
        log.info("Getting Location numbers from Webex")
        if self.calling_enabled is False:
            log.debug("Not a Webex Calling Location")
            return None
        if self._numbers is None:
            self._numbers = NumberList(self)
        return self._numbers

    @property
    def available_numbers(self):
        """Returns all available numbers for the Location.

        Returns both Active and Inactive numbers so that numbers can be assigned prior to activation/porting/

        Returns:
            list[dict]: A list of available numbers, in dict form

        """
        log.debug('Getting available numbers')
        if self.calling_enabled is False:
            log.debug("Not a Webex Calling Location")
            return None
        params = {'locationId': self.id, 'available': True}
        response = webex_api_call("get", "v1/telephony/config/numbers", params=params)
        if len(response.get('phoneNumbers', 0)) == 0:
            return None
        else:
            return response['phoneNumbers']

    @property
    def main_number(self) -> Number:
        """Returns the Main Number for this Location, if defined

        Returns:
            Number: The main number for the Location. None is returned if a main number cannot be found.

        """
        if self._main_number is None:
            config = self.calling_config
            number = self.numbers.get(location=self, phone_number=config['callingLineId'].get('phoneNumber'))
            self._main_number = number
        return self._main_number

    def set_main_number(self, number: Union[Number, str]) -> bool:
        """ Set the Main Number for the Location

        Args:
            number (Number, str): The main number as a :class:`~.number.Number` or a string

        Returns:
            bool: True on success. False otherwise.

        Raises:
            wxcadm.APIError: Raised when the new Main Number is rejected by Webex for any reason

        """
        if isinstance(number, Number):
            number = number.phone_number
        payload = {'callingLineId': {'phoneNumber': number}}
        webex_api_call('put', f"v1/telephony/config/locations/{self.id}", payload=payload,
                       params={'orgId': self.org_id})
        self._main_number = None
        return True

    @property
    def schedules(self):
        """ List of all :class:`wxcadm.LocationSchedule` instances for this Location"""
        log.debug('Getting Location Schedules')
        if self.calling_enabled is False:
            log.debug("Not a Webex Calling Location")
            return None
        response = []
        api_resp = webex_api_call("get", f"v1/telephony/config/locations/{self.id}/schedules", headers=self._headers)
        for schedule in api_resp['schedules']:
            log.debug(f'\tSchedule: {schedule}')
            this_schedule = LocationSchedule(self, schedule['id'], schedule['name'], schedule['type'])
            response.append(this_schedule)
        return response

    @property
    def workspaces(self) -> WorkspaceList:
        """ :class:`WorkspaceList` of all Workspaces for this Location """
        if self._workspaces is None:
            self._workspaces = WorkspaceList(self)
        return self._workspaces

    def set_announcement_language(self, language: str,
                                  update_users: bool = False,
                                  update_features: bool = False) -> bool:
        """ Set the Announcement Language for the Location or update the Announcement Language for existing users
        or features.

        When using this method, setting either ``update_users`` or ``update_features`` will only change the
        Announcement Language for existing users/features. It does not change the Location default that will be used
        for any new users or features. To ensure that the default language is changed, the method should be called
        without those arguments.

        Args:
            language (str): The language code (e.g. ``en_US``) to assign
            update_users (bool, optional): True to update all existing Users and Workspaces to the new language
            update_features (bool, optional): True to update all existing Features to the new language.

        Returns:
            bool: True on success, False otherwise

        """
        log.info(f"Setting language for {self.name} to {language}")
        payload = {
            'announcementLanguageCode': language,
            'agentEnabled': update_users,
            'serviceEnabled': update_features
        }
        success = webex_api_call('post',
                                 f'/v1/telephony/config/locations/{self.id}/actions/modifyAnnouncementLanguage/invoke',
                                 payload=payload)
        log.debug(f"Response: {success}")
        if success:
            log.info("Language change succeeded")
            return True
        else:
            log.info("Language change failed")
            return False

    def upload_moh_file(self, filename: str):
        """ Upload and activate a custom Music On Hold audio file.

        The audio file must be a WAV file that conforms with the Webex Calling requirements. It is recommended to test
        any WAV file by manually uploading it to a Location in Control Hub. The method will return False if the WAV file
        is rejected by Webex.

        Args:
            filename (str): The filename, including path, to the WAV file to upload.

        Returns:
            bool: True on success, False otherwise

        .. warning::

            This method requires the CP-API access scope.

        """
        log.debug(f'Uploading MOH file: {filename}')
        if self.calling_enabled is False:
            log.debug("Not a Webex Calling Location")
            return None
        upload = self._parent._cpapi.upload_moh_file(self.id, filename)
        if upload is True:
            log.debug('\tActivating MOH')
            activate = self._parent._cpapi.set_custom_moh(self.id, filename)
            if activate is True:
                log.debug('\t\tSuccessful activation')
                return True
            else:
                log.debug('\t\tActivation failed')
                return False
        else:
            return False

    def set_default_moh(self):
        """ Set the MOH to be the Webex Calling system default music.

        Returns:
            bool: True on success, False otherwise

        .. warning::

            This method requires the CP-API access scope.
        """
        log.debug('Setting Default MOH')
        if self.calling_enabled is False:
            log.debug("Not a Webex Calling Location")
            return None
        activate = self._parent._cpapi.set_default_moh(self.id)
        if activate is True:
            log.debug('\tSuccessful activation')
            return True
        else:
            log.debug('\tActivation failed')
            return False

    @property
    def outgoing_call_permissions(self):
        """ The Outgoing Call Permissions dicts (as a list) for the Location"""
        if self.calling_enabled is False:
            log.debug("Not a Webex Calling Location")
            return None
        ocp = webex_api_call('get', f'/v1/telephony/config/locations/{self.id}/outgoingPermission')
        return ocp['callingPermissions']

    def set_outgoing_call_permissions(self, outgoing_call_permissions: list) -> Optional[bool]:
        """ Ste the Outgoing Call Permissions for the Location

        This method uses the `callingPermissions` list style of the Webex API, which is the same format as returned by
        the :py:meth:`outgoing_call_permissions` property. The easiest method to change the permissions is to pull the
        `outgoing_call_permissions` list, modify it and pass it back to this method.

        Args:
            outgoing_call_permissions (list): The Webex `callingPermissions` list

        Returns:
            bool: True on success, False otherwise

        """
        log.info(f'Setting Outgoing Call Permission for Location: {self.name}')
        if self.calling_enabled is False:
            log.debug("Not a Webex Calling Location")
            return None
        log.debug(f'\tNew Permissions: {outgoing_call_permissions}')
        if not isinstance(outgoing_call_permissions, list):
            log.warning('outgoing_call_permissions is not a list')
            raise ValueError('outgoing_call_permissions must be a list')

        payload = {'callingPermissions': outgoing_call_permissions}

        success = webex_api_call('put', f'/v1/telephony/config/locations/{self.id}/outgoingPermission',
                                 payload=payload)
        if success:
            return True
        else:
            return False

    @property
    def park_extensions(self):
        """ List of :py:class:`CallParkExtension` instances for this Location """
        log.info(f"Getting Park Extensions for Location: {self.name}")
        if self.calling_enabled is False:
            log.debug("Not a Webex Calling Location")
            return None
        park_extensions = []
        response = webex_api_call("get", "v1/telephony/config/callParkExtensions", params={'locationId': self.id})
        log.debug(f"Response:\n\t{response}")
        for entry in response['callParkExtensions']:
            this_instance = CallParkExtension(self, entry['id'], entry['name'], entry['extension'])
            park_extensions.append(this_instance)

        return park_extensions

    def create_park_extension(self, name: str, extension: str):
        """ Create a new Call Park Extension at the Location

        Args:
            name (str): The name of the Park Extension
            extension (str): The Park Extension

        Returns:
            str: The ID of the newly created CallParkExtension

        """
        log.info(f"Creating Park Extension {name} ({extension}) at Location: {self.name}")
        if self.calling_enabled is False:
            log.debug("Not a Webex Calling Location")
            return None
        payload = {"name": name, "extension": extension}
        response = webex_api_call("post", f"v1/telephony/config/locations/{self.id}/callParkExtensions",
                                  params={'orgId': self._parent.id}, payload=payload)
        log.debug(f"Response:\n\t{response}")
        return response['id']

    @property
    def pickup_groups(self) -> Optional[PickupGroupList]:
        """ :class:`PickupGroupList` list of :py:class:`PickupGroup` for this Location """
        log.info(f"Getting Pickup Groups for Location {self.name}")
        if self.calling_enabled is False:
            log.debug("Not a Webex Calling Location")
            return None
        if self._pickup_groups is None:
            self._pickup_groups = PickupGroupList(self)
        return self._pickup_groups

    @property
    def recording_vendor(self) -> LocationRecordingVendorSelection:
        return LocationRecordingVendorSelection(self)

    @property
    def voice_portal(self) -> VoicePortal:
        """ :class:`VoicePortal` instance for this Location """
        return VoicePortal(self)

    @property
    def dect_networks(self):
        """ :class:`DECTNetworkList` for this Location """
        if self._dect_networks is None:
            self._dect_networks = DECTNetworkList(self)
        return self._dect_networks

    @property
    def pstn(self):
        if self._pstn is None:
            self._pstn = LocationPSTN(self)
        return self._pstn

    @property
    def enhanced_emergency_calling(self) -> LocationEmergencySettings:
        """ The Enhanced Emergency Calling settings for the Location

        Returns:
            LocationEmergencySettings: An object with the boolean attributes ``'integration'`` and ``'routing'``

        """
        response = webex_api_call('get', f"v1/telephony/config/locations/{self.id}/redSky",
                                  params={'orgId': self.org_id})
        return LocationEmergencySettings(integration=response['integrationEnabled'], routing=response['routingEnabled'])

    def set_enhanced_emergency_calling(self, mode: str) -> dict:
        """ Set the Location's Enhanced Emergency Calling mode to one of the following:

        * `'none'` - Locations is opted out of Enhanced Emergency Calling
        * `'integration'` - The Location is integrated with RedSky and sending HELD data and 933 test calls
        * `'routing'` - The Location is routing 911 calls to RedSky

        .. note::
            The Webex APIs exhibit a very strange behavior, so running this method with any values will almost always
            cause the Location to show up in Control Hub with a warning that the compliance status is unknown. The
            Location will be configured correctly but the only way to resolve the warning is by manually completing the
            setup in Control Hub

        Args:
            mode (str): `'none'`, `'integration'`, or `'routing'`

        Returns:
            dict: The `'locationStatus'` dict as returned by Webex

        """
        steps = ['LOCATION_SETUP', 'ALERTS']
        if mode.lower() == 'integration':
            steps.append('NETWORK_ELEMENTS')
        elif mode.lower() == 'routing':
            steps.extend(['NETWORK_ELEMENTS', 'ROUTING_ENABLED', 'COMPLIANT'])
        elif mode.lower() == 'none':
            steps = ['OPTED_OUT']
        else:
            steps = [mode]
        for step in steps:
            payload = {
                'complianceStatus': step
            }
            response = webex_api_call('put', f"v1/telephony/config/locations/{self.id}/redSky/status",
                                      params={'orgId': self.org_id}, payload=payload)
        return response


class LocationFloor:
    """ A Floor within a Location """

    def __init__(self, parent: Location, config: Optional[dict] = None, id: Optional[str] = None):
        self.parent = parent
        if config is None and id is not None:
            config = self._get_data(id)
        self.id: str = config['id']
        self.floor_number: int = config['floorNumber']
        self.name: str = config['displayName']

    def _get_data(self, id: str):
        response = webex_api_call('get', f'v1/locations/{self.parent.id}/floors/{id}')
        return response

    def update(self, floor_number: Optional[int] = None, name: Optional[str] = None) -> bool:
        """ Update the floor number and/or name of the floor

        Args:
            floor_number (int, optional): The new floor number
            name (str, optional): The new name for the floor

        Returns:
            bool: True on success

        """
        payload = {}
        if floor_number is not None:
            payload['floorNumber'] = floor_number
        if name is not None:
            payload['displayName'] = name
        response = webex_api_call('put', f'v1/locations/{self.parent.id}/floors/{self.id}', payload=payload,
                                  params={'orgId': self.parent.org_id})
        self.floor_number = response['floorNumber']
        self.name = response['displayName']
        return True

    def delete(self) -> bool:
        """ Delete the floor

        Returns:
            bool: True on success

        """
        webex_api_call('delete', f'v1/locations/{self.parent.id}/floors/{self.id}',
                       params={'orgId': self.parent.org_id})
        return True


class LocationFloorList(UserList):
    def __init__(self, parent: Location):
        super().__init__()
        """ List of :class:`LocationFloor` instances for the :class:`Location` """
        self.parent = parent
        self.data: list = []

        self._get_data()

    def _get_data(self):
        self.data = []
        response = webex_api_call('get', f'v1/locations/{self.parent.id}/floors',
                                  params={'orgId': self.parent.org_id})
        for floor in response:
            self.data.append(LocationFloor(self.parent, floor))

    def refresh(self):
        self._get_data()

    def create(self, floor_number: int, name: str):
        """ Create a new floor in the Location

        Args:
            floor_number (int): The floor number within the building
            name (str): The descriptive name of the floor (e.g. 'Basement', '2nd Floor')

        Returns:
            LocationFloor: The :class:`LocationFloor` instance that was created.

        Raises:
            ValueError: Raised when the ``floor_number`` already exists

        """
        # Validate the floor number doesn't exist
        for floor in self.data:
            if floor.floor_number == floor_number:
                raise ValueError('Duplicate floor number')
        payload = {
            'floorNumber': floor_number,
            'displayName': name
        }
        response = webex_api_call('post', f'v1/locations/{self.parent.id}/floors', payload=payload,
                                  params={'orgId': self.parent.org_id})
        self.data.append(LocationFloor(self.parent, config=response))



from builtins import Exception

__all__ = ['OrgError', 'LicenseError', 'APIError', 'TokenError', 'PutError', 'XSIError', 'NotAllowed', 'CSDMError']


class OrgError(Exception):
    """The base error class for Org-related exceptions"""
    def __init__(self, message):
        super(OrgError, self).__init__(message)


class LicenseError(OrgError):
    """Exceptions dealing with License problems within the Org"""
    def __init__(self, message):
        super(LicenseError, self).__init__(message)


class APIError(Exception):
    """The base class for any exceptions dealing with the API"""
    def __init__(self, message):
        super(APIError, self).__init__(message)


class TokenError(APIError):
    """Exceptions dealing with the Access Token itself"""
    def __init__(self, message):
        super(TokenError, self).__init__(message)


class PutError(APIError):
    """Exception class for problems putting values back into Webex"""
    pass


class XSIError(APIError):
    """Exception class for problems with the XSI API. Serves as a base class for other errors."""
    pass


class NotAllowed(XSIError):
    """Exception class for XSI actions that are not allowed by the platform due to user settings"""
    pass


class CSDMError(APIError):
    def __init__(self, message):
        super().__init__(message)



from __future__ import annotations

from typing import Optional, Union
from collections import UserList

import wxcadm.exceptions
from .common import *
from wxcadm import log


class HuntGroup:
    def __init__(self, parent: wxcadm.Org | wxcadm.Location, id: str, config: Optional[dict] = None):
        """Initialize a HuntGroup instance

        .. note::
            The HuntGroup class is normally not instantiated manually. It is created by the :class:`HuntGroupList`.

        Args:
            parent (Location|Org): The Location or Org instance to which the Hunt Group belongs
            id (str): The Webex ID for the Hunt Group
            config (dict, optional): The full config dict from Webex

        Returns:
            HuntGroup: The HuntGroup instance

        """
        log.debug(f"Initializing HuntGroup instance")

        # Instance attrs
        self.parent: wxcadm.Org | wxcadm.Location = parent
        self.id: str = id
        """The Webex ID of the Hunt Group"""
        self.name: str = config.get('name', '')
        """The name of the Hunt Group"""
        self.location_id: str = config['locationId']
        """The Location ID associated with the Hunt Group"""
        self.enabled: bool = config.get('enabled', True)
        """Whether the Hunt Group is enabled or not"""
        self.phone_number: str = config.get('phoneNumber', '')
        """The DID for the Hunt Group"""
        self.extension: str = config.get('extension', '')
        """The extension of the Hunt Group"""

    def __str__(self):
        return self.name

    def __repr__(self):
        return self.id

    @property
    def config(self) -> dict:
        """ The config of the Hunt Group """
        response = webex_api_call("get", f"v1/telephony/config/locations/{self.location_id}/huntGroups/{self.id}")
        return response

    @property
    def agents(self) -> list:
        return self.config['agents']

    @property
    def spark_id(self) -> str:
        return decode_spark_id(self.id)

    def add_agent(self, agent: Union[wxcadm.Person, wxcadm.Workspace, wxcadm.VirtualLine],
                  weight: Optional[str] = None) -> bool:
        """ Add an agent to the Hunt Group

        The weight field is only applicable and required when the call policy is `'WEIGHTED'`

        Args:
            agent: The :class:`Person`, :class:`Workspace` or :class:`VirtualLine` to add as an agent
            weight: The weight to associate with the agent

        Returns:
            bool: True on success

        """
        log.info(f"Adding agent to Hunt Group {self.name}")
        config = self.config
        new_agent_payload = {'id': agent.id}
        if weight is not None:
            new_agent_payload['weight'] = weight
        config['agents'].append(new_agent_payload)
        webex_api_call('put', f"v1/telephony/config/locations/{self.location_id}/huntGroups/{self.id}",
                       payload=config)
        return True


class HuntGroupList(UserList):
    _endpoint = "v1/telephony/config/huntGroups"
    _endpoint_items_key = "huntGroups"
    _item_endpoint = "v1/telephony/config/locations/{location_id}/huntGroups/{item_id}"
    _item_class = HuntGroup

    def __init__(self, parent: Union[wxcadm.Org, wxcadm.Location]):
        super().__init__()
        log.debug("Initializing HuntGroupList")
        self.parent: wxcadm.Org | wxcadm.Location = parent
        self.data: list = self._get_data()

    def _get_data(self) -> list:
        log.debug("_get_data() started")
        params = {}

        if isinstance(self.parent, wxcadm.Org):
            log.debug(f"Using Org ID {self.parent.id} as Data filter")
            params['orgId'] = self.parent.id
        elif isinstance(self.parent, wxcadm.Location):
            log.debug(f"Using Location ID {self.parent.id} as Data filter")
            params['locationId'] = self.parent.id
        else:
            log.warn("Parent class is not Org or Location, so all items will be returned")
        response = webex_api_call('get', self._endpoint, params=params)
        log.info(f"Found {len(response[self._endpoint_items_key])} items")

        items = []
        for entry in response[self._endpoint_items_key]:
            items.append(self._item_class(parent=self.parent, id=entry['id'], config=entry))
        return items

    def refresh(self):
        """ Refresh the list of instances from Webex

        Returns:
            bool: True on success, False otherwise.

        """
        self.data = self._get_data()
        return True

    def get(self, id: Optional[str] = None, name: Optional[str] = None, spark_id: Optional[str] = None,
            uuid: Optional[str] = None):
        """ Get the instance associated with a given ID, Name, or Spark ID

        Only one parameter should be supplied in normal cases. If multiple arguments are provided, the Locations will be
        searched in order by ID, Name, UUID and Spark ID. If no arguments are provided, the method will raise an
        Exception.

        Args:
            id (str, optional): The Call Queue ID to find
            name (str, optional): The Call Queue Name to find. Case-insensitive.
            spark_id (str, optional): The Spark ID to find
            uuid (str, optional): The UUID to find

        Returns:
            CallQueue: The CallQueue instance correlating to the given search argument.

        Raises:
            ValueError: Raised when the method is called with no arguments

        """
        if id is None and name is None and spark_id is None:
            raise ValueError("A search argument must be provided")
        if id is not None:
            for item in self.data:
                if item.id == id:
                    return item
        if name is not None:
            for item in self.data:
                if item.name.lower() == name.lower():
                    return item
        if uuid is not None:
            for item in self.data:
                if item.spark_id.split('/')[-1].uppper() == uuid.upper():
                    return item
        if spark_id is not None:
            for item in self.data:
                if item.spark_id == spark_id:
                    return item
        return None

    def create(self,
               name: str,
               first_name: str,
               last_name: str,
               phone_number: Optional[str] = None,
               extension: Optional[str] = None,
               call_policies: Optional[dict] = None,
               enabled: bool = True,
               language: Optional[str] = None,
               time_zone: Optional[str] = None,
               location: Optional[wxcadm.Location] = None,
               agents: Optional[list] = None,
               allow_as_agent_caller_id: bool = False
               ):
        """ Create a Hunt Group

        Args:
            name (str): The name of the Hunt Group
            first_name (str): The first name to be used for Caller ID
            last_name (str): The last name to be used for Caller ID
            phone_number (str, optional): The phone number (e.g. DID) for the Hunt Group
            extension (str, optional): The extension for the Hunt Group
            call_policies (dict, optional): The Hunt Group Call Policy. The dict has the following format::

                {
                    'policy': 'REGULAR',
                    'waitingEnabled': True,
                    'noAnswer': {
                        'nextAgentEnabled': False,
                        'nextAgentRings': 2,
                        'forwardEnabled': True,
                        'destination': '5050',
                        'numberOfRings': 3,
                        'systemMaxNumberOfRings': 20,
                        'destinationVoicemailEnabled': False
                    },
                    'businessContinuity': {
                        'enabled': False,
                        'destination': '6312128008',
                        'destinationVoicemailEnabled': False
                    }
                }

            enabled (bool, optional): Whether the Hunt Group is enabled. Defaults to True
            language (str, optional): The language for audio prompts. Defaults to ``Location.announcement_language``
            time_zone (str, optional): The time zone of the Hunt Group. Defaults to ``Location.time_zone``
            location (Location, optional): The Location at which to create the Hunt Group. Only required when the
                :class:`HuntGroupList` is at the Org level. If at the Location level, the selected :class:`Location`
                will be used.
            agents: (list, optional): A list of :class:`Person`, :class:`Workspace` or :class:`VirtualLine` instances
                which will be assigned as agents in the Hunt Group. Without this list, no agents will be present.
            allow_as_agent_caller_id (bool, optional): Whether the Hunt Group phone number can be used as Caller ID
                by agents. Defaults to False.

        Returns:
            HuntGroup: The :class:`HuntGroup` instance of the created Hunt Group

        """
        if location is None and isinstance(self.parent, wxcadm.Org):
            raise ValueError("location is required for Org-level HuntGroupList")
        elif location is None and isinstance(self.parent, wxcadm.Location):
            location = self.parent
        log.info(f"Creating Hunt Group at Location {location.name} with name: {name}")
        if location.calling_enabled is False:
            log.debug("Not a Webex Calling Location")
            return None
        # Get some values if they weren't passed
        if phone_number is None and extension is None:
            raise ValueError("phone_number and/or extension are required")
        if time_zone is None:
            log.debug(f"Using Location time_zone {location.time_zone}")
            time_zone = location.time_zone
        if language is None:
            log.debug(f"Using Location announcement_language {location.announcement_language}")
            language = location.announcement_language
        if call_policies is None:
            log.debug(f"Using default Call Policies")
            call_policies = {
                'policy': 'SIMULTANEOUS',
                'waitingEnabled': False,
                'groupBusyEnabled': False,
                'allowMembersToControlGroupBusyEnabled': False,
                'noAnswer': {
                    'nextAgentEnabled': False,
                    'nextAgentRings': 5,
                    'forwardEnabled': False,
                    'numberOfRings': 15,
                    'systemMaxNumberOfRings': 20,
                    'destinationVoicemailEnabled': False
                },
                'busyRedirect': {
                    'enabled': False,
                    'destinationVoicemailEnabled': False,
                },
                'businessContinuityRedirect': {
                    'enabled': False,
                    'destinationVoicemailEnabled': False
                }
            }

        payload = {
            "name": name,
            "firstName": first_name,
            "lastName": last_name,
            "extension": extension,
            "phoneNumber": phone_number,
            "timeZone": time_zone,
            "languageCode": language,
            "callPolicies": call_policies,
            'huntGroupCallerIdForOutgoingCallsEnabled': allow_as_agent_caller_id
        }
        if agents is not None:
            agent_list = []
            log.info("Finding agent IDs to assign to Hunt Group")
            for agent in agents:
                if not isinstance(agent, wxcadm.Person) and not isinstance(agent, wxcadm.VirtualLine) and \
                        not isinstance(agent, wxcadm.Workspace):
                    raise ValueError("Agents must be of type Workspace, Person or VirtualLine")
                agent_list.append({'id': agent.id})
            payload['agents'] = agent_list

        response = webex_api_call("post", f"v1/telephony/config/locations/{location.id}/huntGroups",
                                  payload=payload)
        new_hg_id = response['id']
        self.refresh()
        return self.get(id=new_hg_id)
